from basinmaker.func.arcgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *


def delineate_watershed_no_lake_using_dem(
    grassdb,
    grass_location,
    qgis_prefix_path,
    input_geo_names,
    acc_thresold,
    fdr_arcgis,
    fdr_grass,
    str_r,
    str_v,
    acc,
    cat_no_lake,
    max_memroy,
):
    work_folder = grassdb
    mask = input_geo_names["mask"]
    dem = input_geo_names["dem"]

    # read and define arcgis work enviroments
    if not os.path.exists(work_folder):
        os.makedirs(work_folder)
    arcpy.env.workspace = work_folder
     
    arcpy.env.overwriteOutput = True
    arcpy.CheckOutExtension("Spatial")
    cellSize = float(arcpy.GetRasterProperties_management(dem+'.tif', "CELLSIZEX").getOutput(0))
    SptailRef = arcpy.Describe(dem+'.tif').spatialReference
    arcpy.env.XYTolerance = cellSize
    arcpy.arcpy.env.cellSize = cellSize    
    arcpy.env.outputCoordinateSystem = arcpy.SpatialReference(int(SptailRef.factoryCode)) ### WGS84
    arcpy.env.extent = arcpy.Describe(dem + '.tif').extent

    outFlowDirection = FlowDirection(dem + '.tif', "NORMAL")
    dirraster = SetNull(Raster(outFlowDirection) < 1, Raster(outFlowDirection))
    dirraster.save(fdr_arcgis + '.tif')

    outFlowAccumulation = FlowAccumulation(dirraster)
    outFlowAccumulation.save(acc + ".tif")

    StreamRaster = SetNull(Raster(outFlowAccumulation) < acc_thresold, Raster(outFlowAccumulation))
    StreamRaster = Con(StreamRaster >= 0, 1, 0)
    StreamRaster.save("str_1.tif")
    
    Strlink = StreamLink(StreamRaster, dirraster)
    Strlink.save(str_r + '.tif')
    
    Catchment = Watershed(dirraster,Strlink)

    Catchment.save(cat_no_lake+'.tif')

    StreamToFeature(Strlink, dirraster, str_v + '.shp',"NO_SIMPLIFY")
    
    # copyfile( OutputFolder + "/"+"dir.prj" ,  OutputFolder + "/"+"Cat1.prj")
    # StreamToFeature(Strlink, dirraster, "DrainL1","NO_SIMPLIFY")
    # copyfile( OutputFolder + "/"+"HyMask.prj" ,  OutputFolder + "/"+"DrainL1.prj")
    # arcpy.RasterToPolygon_conversion("Cat1", "Cattemp.shp", "NO_SIMPLIFY")
    # copyfile( OutputFolder + "/"+"HyMask.prj" ,  OutputFolder + "/"+"Cattemp.prj")
    # arcpy.Dissolve_management("Cattemp.shp", "Cat1.shp", ["gridcode"])
    # copyfile( OutputFolder + "/"+"HyMask.prj" ,  OutputFolder + "/"+"Cat1.prj")
    # arcpy.AddField_management("Cat1.shp", "COAST", "LONG",10,"","", "", "NULLABLE")
    # arcpy.AddField_management("Cat1.shp", "HYBAS_ID", "LONG",10,"","", "", "NULLABLE")
    # arcpy.AddField_management("Cat1.shp", "NEXT_DOWN", "LONG",10,"","", "", "NULLABLE")
    # fieldList = ["GRID_CODE", "FROM_NODE","TO_NODE"]
    # arcpy.JoinField_management("Cat1.shp", "gridcode", "DrainL1.shp", "GRID_CODE", fieldList)
    # arcpy.CalculateField_management("Cat1.shp", "HYBAS_ID", '!FROM_NODE! * 1', "PYTHON")
    # arcpy.CalculateField_management("Cat1.shp", "NEXT_DOWN", '!TO_NODE! * 1', "PYTHON")

    return



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
from basinmaker.delineationnolake.watdelineationwithoutlake import (
    watershed_delineation_without_lake,
)
from basinmaker.addlakeandobs.addlakesqgis import (
    add_lakes_into_existing_watershed_delineation,
)
from basinmaker.addattributes.calbkfwidthdepthqgis import (
    calculate_bankfull_width_depth_from_polyline,
)
import tempfile
import sqlite3


def Generatesubdomain(
    input_geo_names,
    grassdb,
    grass_location,
    qgis_prefix_path,
    path_lakefile_in,
    lake_attributes,
    Min_Num_Domain=9,
    Max_Num_Domain=13,
    Initaial_Acc=5000,
    Delta_Acc=1000,
    CheckLakeArea=1,
    fdr_path="#",
    Acc_Thresthold_stream=500,
    max_memory=2048 * 3,
    Out_Sub_Reg_Folder="#",
    sub_reg_str_r="sub_reg_str_r",
    sub_reg_str_v="sub_reg_str_v",
    sub_reg_nfdr_grass="sub_reg_nfdr_grass",
    sub_reg_nfdr_arcgis="sub_reg_nfdr_arcgis",
    sub_reg_acc="sub_reg_acc",
    sub_reg_dem="sub_reg_dem",
    cat_add_lake="cat_add_lake",
):

    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    if not os.path.exists(Out_Sub_Reg_Folder):
        os.makedirs(Out_Sub_Reg_Folder)

    # required inputs
    dem = input_geo_names["dem"]
    mask = input_geo_names["mask"]

    # define local variable file names
    fdr_arcgis = Internal_Constant_Names["fdr_arcgis"]
    fdr_grass = Internal_Constant_Names["fdr_grass"]
    nfdr_arcgis = Internal_Constant_Names["nfdr_arcgis"]
    nfdr_grass = Internal_Constant_Names["nfdr_grass"]
    str_r = Internal_Constant_Names["str_r"]
    str_v = Internal_Constant_Names["str_v"]
    acc = Internal_Constant_Names["acc"]
    cat_no_lake = Internal_Constant_Names["cat_no_lake"]
    sl_connected_lake = Internal_Constant_Names["sl_connected_lake"]
    sl_non_connected_lake = Internal_Constant_Names["sl_nonconnect_lake"]
    sl_lakes = Internal_Constant_Names["selected_lakes"]
    catchment_without_merging_lakes = Internal_Constant_Names[
        "catchment_without_merging_lakes"
    ]
    river_without_merging_lakes = Internal_Constant_Names["river_without_merging_lakes"]
    cat_use_default_acc = Internal_Constant_Names["cat_use_default_acc"]
    pourpoints_with_lakes = Internal_Constant_Names["pourpoints_with_lakes"]
    pourpoints_add_obs = Internal_Constant_Names["pourpoints_add_obs"]
    lake_outflow_pourpoints = Internal_Constant_Names["lake_outflow_pourpoints"]
    all_lakes = input_geo_names["all_lakes"]
    lake_boundary = input_geo_names["lake_boundary"]

    #### Determine Sub subregion without lake
    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
    )
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")
    N_Basin = 0
    Acc = Initaial_Acc
    print("##############################Loop for suitable ACC ")
    while N_Basin < Min_Num_Domain or N_Basin > Max_Num_Domain:
        grass.run_command(
            "r.watershed",
            elevation=dem,
            flags="sa",
            basin="testbasin",
            drainage="dir_grass_reg",
            accumulation="acc_grass_reg2",
            threshold=Acc,
            overwrite=True,
        )
        N_Basin, temp = generate_stats_list_from_grass_raster(
            grass, mode=1, input_a="testbasin"
        )
        N_Basin = np.unique(N_Basin)
        N_Basin = len(N_Basin[N_Basin > 0])
        print(
            "Number of Subbasin:    ",
            N_Basin,
            "Acc  value:     ",
            Acc,
            "Change of ACC ",
            Delta_Acc,
        )
        if N_Basin > Max_Num_Domain:
            Acc = Acc + Delta_Acc
        if N_Basin < Min_Num_Domain:
            Acc = Acc - Delta_Acc
    PERMANENT.close()

    if fdr_path == "#":
        mode = "usingdem"
    else:
        mode = "usingfdr"

    watershed_delineation_without_lake(
        mode=mode,
        input_geo_names=input_geo_names,
        acc_thresold=Acc,
        fdr_path=fdr_path,
        fdr_arcgis=fdr_arcgis,
        fdr_grass=fdr_grass,
        str_r=str_r,
        str_v=str_v,
        acc=acc,
        cat_no_lake=cat_no_lake,
        max_memroy=max_memory,
        grassdb=grassdb,
        grass_location=grass_location,
        qgis_prefix_path=qgis_prefix_path,
        gis_platform="qgis",
    )
    input_geo_names["fdr_arcgis"] = "fdr_arcgis"
    input_geo_names["fdr_grass"] = "fdr_grass"
    input_geo_names["str_r"] = "str_r"
    input_geo_names["str_v"] = "str_v"
    input_geo_names["acc"] = "acc"
    input_geo_names["cat_no_lake"] = "cat_no_lake"

    add_lakes_into_existing_watershed_delineation(
        grassdb=grassdb,
        grass_location=grass_location,
        qgis_prefix_path=qgis_prefix_path,
        input_geo_names=input_geo_names,
        path_lakefile_in=path_lakefile_in,
        lake_attributes=lake_attributes,
        threshold_con_lake=CheckLakeArea,
        threshold_non_con_lake=100000,
        remove_lake_inlets = True,
        only_included_lake_at_river_interction = True,
        sl_connected_lake=sl_connected_lake,
        sl_non_connected_lake=sl_non_connected_lake,
        sl_lakes=sl_lakes,
        nfdr_arcgis=nfdr_arcgis,
        nfdr_grass=nfdr_grass,
        cat_add_lake=cat_add_lake,
        pourpoints_with_lakes=pourpoints_with_lakes,
        cat_use_default_acc=cat_use_default_acc,
        lake_outflow_pourpoints=lake_outflow_pourpoints,
        max_memroy=max_memory,
    )

    ####Determin river network for whole watersheds
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")
    grass.run_command(
        "r.stream.extract",
        elevation=dem,
        accumulation=acc,
        threshold=Acc_Thresthold_stream,
        stream_raster=sub_reg_str_r,
        stream_vector=sub_reg_str_v,
        overwrite=True,
        memory=max_memory,
    )

    #### export outputs
    grass.run_command(
        "r.pack",
        input=nfdr_grass,
        output=os.path.join(Out_Sub_Reg_Folder, sub_reg_nfdr_grass + ".pack"),
        overwrite=True,
    )
    grass.run_command(
        "r.pack",
        input=nfdr_arcgis,
        output=os.path.join(Out_Sub_Reg_Folder, sub_reg_nfdr_arcgis + ".pack"),
        overwrite=True,
    )
    grass.run_command(
        "r.pack",
        input=fdr_grass,
        output=os.path.join(Out_Sub_Reg_Folder, 'sub_reg_fdr_grass' + ".pack"),
        overwrite=True,
    )
    grass.run_command(
        "r.pack",
        input=fdr_arcgis,
        output=os.path.join(Out_Sub_Reg_Folder, 'sub_reg_fdr_arcgis' + ".pack"),
        overwrite=True,
    )
    grass.run_command(
        "r.pack",
        input=acc,
        output=os.path.join(Out_Sub_Reg_Folder, sub_reg_acc + ".pack"),
        overwrite=True,
    )
    grass.run_command(
        "r.pack",
        input=input_geo_names["dem"],
        output=os.path.join(Out_Sub_Reg_Folder, sub_reg_dem + ".pack"),
        overwrite=True,
    )
    grass.run_command(
        "v.pack",
        input=sub_reg_str_v,
        output=os.path.join(Out_Sub_Reg_Folder, sub_reg_str_v + ".pack"),
        overwrite=True,
    )
    grass.run_command(
        "r.pack",
        input=sub_reg_str_r,
        output=os.path.join(Out_Sub_Reg_Folder, sub_reg_str_r + ".pack"),
        overwrite=True,
    )

    grass.run_command(
        "r.pack",
        input=all_lakes,
        output=os.path.join(Out_Sub_Reg_Folder, all_lakes + ".pack"),
        overwrite=True,
    )
    grass.run_command(
        "r.pack",
        input=lake_boundary,
        output=os.path.join(Out_Sub_Reg_Folder, lake_boundary + ".pack"),
        overwrite=True,
    )

    PERMANENT.close()
    return


def generatesubdomainmaskandinfo(
    Out_Sub_Reg_Dem_Folder,
    input_geo_names,
    grassdb,
    grass_location,
    qgis_prefix_path,
    path_bkfwidthdepth,
    bkfwd_attributes,
):
    ###
    dem = input_geo_names["dem"]
    cat_add_lake = input_geo_names["cat_add_lake"]
    ndir_Arcgis = input_geo_names["nfdr_arcgis"]
    acc_grass = input_geo_names["acc"]
    str_r = input_geo_names["str_r"]
    outlet_pt_info = "outlet_pt_info"

    maximum_obs_id = 80000
    tempfolder = os.path.join(
        tempfile.gettempdir(),
        "basinmaker_subreg" + str(np.random.randint(1, 10000 + 1)),
    )
    if not os.path.exists(tempfolder):
        os.makedirs(tempfolder)

    k = -1
    c = -1
    if path_bkfwidthdepth != "#":
        k, c = calculate_bankfull_width_depth_from_polyline(
            grassdb=grassdb,
            grass_location=grass_location,
            qgis_prefix_path=qgis_prefix_path,
            path_bkfwidthdepth=path_bkfwidthdepth,
            bkfwd_attributes=bkfwd_attributes,
            catinfo=[],
            input_geo_names=input_geo_names,
            k_in=-1,
            c_in=-1,
            return_k_c_only=True,
        )

    #### generate subbregion outlet points and subregion info table
    QgsApplication.setPrefixPath(qgis_prefix_path, True)
    Qgs = QgsApplication([], False)
    Qgs.initQgis()
    from processing.core.Processing import Processing
    from processing.tools import dataobjects
    from qgis import processing

    feedback = QgsProcessingFeedback()
    Processing.initialize()
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    context = dataobjects.createContext()
    context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)

    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="-1")
    )
    con = sqlite3.connect(
        os.path.join(grassdb, grass_location, "PERMANENT", "sqlite", "sqlite.db")
    )

    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")
    grass.run_command("r.mask", raster=dem, maskcats="*", overwrite=True)
    grass.run_command("r.null", map=cat_add_lake, setnull=-9999)

    exp = "%s = if(isnull(%s),null(),%s)" % (
        "river_r",
        str_r,
        cat_add_lake,
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    routing_temp = generate_routing_info_of_catchments(
        grass,
        con,
        cat=cat_add_lake,
        acc=acc_grass,
        Name="Final",
        str="river_r",
        garray=garray,
    )

    grass.run_command("g.copy", vector=("Final_OL_v", outlet_pt_info), overwrite=True)
    grass.run_command(
        "g.copy", vector=("Final_IL_v_c", "sub_reg_inlet"), overwrite=True
    )
    Paths_Finalcat_ply = []

    sqlstat = (
        "SELECT SubId, DowSubId,ILSubIdmax,ILSubIdmin,MaxAcc_cat,ILpt_ID FROM %s"
        % (outlet_pt_info)
    )
    outletinfo = pd.read_sql_query(sqlstat, con)
    outletinfo = outletinfo.fillna(-1)
    outletinfo = outletinfo.loc[outletinfo["SubId"] > 0]

    sqlstat = "SELECT ILpt_ID,SubId_I FROM %s" % ("Final_IL_v_c")
    inletinfo = pd.read_sql_query(sqlstat, con)
    inletinfo = inletinfo.fillna(-1)

    # update watershed bankfull k and c first
    outletinfo["k"] = k
    outletinfo["c"] = c

    subregin_info = pd.DataFrame(
        np.full(len(outletinfo), -9999), columns=["Sub_Reg_ID"]
    )
    subregin_info["Dow_Sub_Reg_Id"] = -9999
    subregin_info["ProjectNM"] = -9999
    subregin_info["Nun_Grids"] = -9999
    subregin_info["Ply_Name"] = -9999
    subregin_info["Max_ACC"] = -9999
    subregin_info["ILpt_ID"] = -9999

    for i in range(0, len(outletinfo)):

        basinid = int(outletinfo["SubId"].values[i])

        grass.run_command("r.mask", raster=dem, maskcats="*", overwrite=True)
        exp = "%s = if(%s == %s,%s,null())" % (
            "dem_reg_" + str(basinid),
            cat_add_lake,
            str(basinid),
            dem,
        )
        grass.run_command("r.mapcalc", expression=exp, overwrite=True)
        ####define mask
        grass.run_command(
            "r.mask", raster="dem_reg_" + str(basinid), maskcats="*", overwrite=True
        )
        grass.run_command(
            "r.out.gdal",
            input="MASK",
            output=os.path.join(tempfolder, "Mask1.tif"),
            format="GTiff",
            overwrite=True,
        )
        processing.run(
            "gdal:polygonize",
            {
                "INPUT": os.path.join(tempfolder, "Mask1.tif"),
                "BAND": 1,
                "FIELD": "DN",
                "EIGHT_CONNECTEDNESS": False,
                "EXTRA": "",
                "OUTPUT": os.path.join(
                    tempfolder, "HyMask_region_" + str(basinid) + ".shp"
                ),
            },
        )
        processing.run(
            "gdal:dissolve",
            {
                "INPUT": os.path.join(
                    tempfolder, "HyMask_region_" + str(basinid) + ".shp"
                ),
                "FIELD": "DN",
                "OUTPUT": os.path.join(
                    tempfolder, "HyMask_region_f" + str(basinid) + ".shp"
                ),
            },
        )
        processing.run(
            "gdal:dissolve",
            {
                "INPUT": os.path.join(
                    tempfolder, "HyMask_region_" + str(basinid) + ".shp"
                ),
                "FIELD": "DN",
                "OUTPUT": os.path.join(
                    Out_Sub_Reg_Dem_Folder,
                    "HyMask_region_"
                    + str(int(basinid + maximum_obs_id))
                    + "_nobuffer.shp",
                ),
            },
        )

        Paths_Finalcat_ply.append(
            os.path.join(
                Out_Sub_Reg_Dem_Folder,
                "HyMask_region_" + str(int(basinid + maximum_obs_id)) + "_nobuffer.shp",
            )
        )

        processing.run(
            "native:buffer",
            {
                "INPUT": os.path.join(
                    tempfolder, "HyMask_region_f" + str(basinid) + ".shp"
                ),
                "DISTANCE": 0.005,
                "SEGMENTS": 5,
                "END_CAP_STYLE": 0,
                "JOIN_STYLE": 0,
                "MITER_LIMIT": 2,
                "DISSOLVE": True,
                "OUTPUT": os.path.join(
                    Out_Sub_Reg_Dem_Folder,
                    "HyMask_region_" + str(int(basinid + maximum_obs_id)) + ".shp",
                ),
            },
        )

    qgis_vector_merge_vector_layers(
        processing,
        context,
        INPUT_Layer_List=Paths_Finalcat_ply,
        OUTPUT=os.path.join(Out_Sub_Reg_Dem_Folder, "subregion_ply.shp"),
    )

    grass.run_command("r.mask", raster=dem, maskcats="*", overwrite=True)

    problem_subid = []
    for i in range(0, len(outletinfo)):
        basinid = int(outletinfo["SubId"].values[i])
        dowsubreginid = int(outletinfo["DowSubId"].values[i])
        ILpt_ID = outletinfo["ILpt_ID"].values[i]
        subregin_info.loc[i, "ILpt_ID"] = ILpt_ID
        if len(inletinfo[inletinfo["ILpt_ID"] == ILpt_ID]["SubId_I"]) > 0:
            downsubid_inlet = inletinfo[inletinfo["ILpt_ID"] == ILpt_ID][
                "SubId_I"
            ].values[0]
            if dowsubreginid != downsubid_inlet:
                problem_subid.append(basinid)

        catacc = int(outletinfo["MaxAcc_cat"].values[i])

        subregin_info.loc[i, "ProjectNM"] = (
            "sub_reg" + "_" + str(int(basinid + maximum_obs_id))
        )
        subregin_info.loc[i, "Ply_Name"] = (
            "HyMask_region_" + str(int(basinid + maximum_obs_id)) + ".shp"
        )
        subregin_info.loc[i, "Max_ACC"] = catacc

        if basinid == dowsubreginid:
            subregin_info.loc[i, "Dow_Sub_Reg_Id"] = int(-1 + maximum_obs_id)
        else:
            subregin_info.loc[i, "Dow_Sub_Reg_Id"] = int(dowsubreginid + maximum_obs_id)
        subregin_info.loc[i, "Sub_Reg_ID"] = int(basinid + maximum_obs_id)

    subregin_info["k"] = k
    subregin_info["c"] = c

    ### remove subregion do not contribute to the outlet
    ## find watershed outlet subregion
    #    subregin_info  = subregin_info.loc[subregin_info['Dow_Sub_Reg_Id'] == self.maximum_obs_id-1]
    subregin_info = subregin_info.sort_values(by="Max_ACC", ascending=False)
    outlet_reg_id = subregin_info["Sub_Reg_ID"].values[0]
    routing_info = (
        subregin_info[["Sub_Reg_ID", "Dow_Sub_Reg_Id"]].astype("float").values
    )
    needed_sub_reg_ids = defcat(routing_info, outlet_reg_id)

    mask = subregin_info["Sub_Reg_ID"].isin(needed_sub_reg_ids)
    subregin_info = subregin_info.loc[mask, :]
    #        subregin_info.drop(subregin_info.index[del_row_mask]) ###
    subregin_info.to_csv(
        os.path.join(Out_Sub_Reg_Dem_Folder, "Sub_reg_info.csv"),
        index=None,
        header=True,
    )

    subregin_info.to_csv(
        os.path.join(Out_Sub_Reg_Dem_Folder, "Sub_reg_info.csv"),
        index=None,
        header=True,
    )

    grass.run_command("v.db.addcolumn", map=outlet_pt_info, columns="reg_subid int")
    grass.run_command("v.db.addcolumn", map=outlet_pt_info, columns="reg_dowid int")

    grass.run_command("v.db.addcolumn", map=outlet_pt_info, columns="sub_reg_id int")

    grass.run_command(
        "v.db.update",
        map=outlet_pt_info,
        column="reg_subid",
        qcol="SubId + " + str(maximum_obs_id),
    )

    grass.run_command(
        "v.db.update",
        map=outlet_pt_info,
        column="sub_reg_id",
        qcol="SubId + " + str(maximum_obs_id),
    )

    grass.run_command(
        "v.db.update",
        map=outlet_pt_info,
        column="reg_dowid",
        qcol="DowSubId + " + str(maximum_obs_id),
    )

    grass.run_command("v.db.addcolumn", map="sub_reg_inlet", columns="sub_reg_id int")

    grass.run_command(
        "v.db.update",
        map="sub_reg_inlet",
        column="sub_reg_id",
        qcol="SubId_I + " + str(maximum_obs_id),
    )

    grass.run_command(
        "v.out.ogr",
        input=outlet_pt_info,
        output=os.path.join(Out_Sub_Reg_Dem_Folder, outlet_pt_info + ".shp"),
        format="ESRI_Shapefile",
        overwrite=True,
    )

    grass.run_command(
        "v.out.ogr",
        input="sub_reg_inlet",
        output=os.path.join(Out_Sub_Reg_Dem_Folder, "sub_reg_inlet" + ".shp"),
        format="ESRI_Shapefile",
        overwrite=True,
    )

    grass.run_command(
        "v.pack",
        input=outlet_pt_info,
        output=os.path.join(Out_Sub_Reg_Dem_Folder, "Sub_Reg_Outlet_v" + ".pack"),
        overwrite=True,
    )
    grass.run_command(
        "r.pack",
        input="Final_OL",
        output=os.path.join(Out_Sub_Reg_Dem_Folder, "Sub_Reg_Outlet_r" + ".pack"),
        overwrite=True,
    )

    print("following subregion's inlet needs to be checked ")
    print(problem_subid)

    return


############################################################################


def Combine_Sub_Region_Results(
    Sub_Region_info,
    Sub_Region_OutputFolder,
    OutputFolder,
    Is_Final_Result,
    qgis_prefix_path,
    subregion_inlet,
    start_sub_id = 0,
    k = 1,
    c = 1,
):
    """Combine subregion watershed delineation results

    It is a function that will combine watershed delineation results
    in different subregions. This function will assgin new subbasin
    id to each polygon in the combined result and update the
    stream orders in the combined result

    Parameters
    ----------
    Sub_Region_info                   : string
        It is the path to a csv file that contains the subregion
        information, such as subregion id etc. It is the output of
        Generatesubdomainmaskandinfo
    Sub_Region_OutputFolder           : string
        It is the path to the subregion output folder.
    OutputFolder                      : string
        It is the path to a folder to save outputs
    Is_Final_Result                   : bool
       Indicate the function is called to combine subbasin polygon of
       final delineation results or subbasin polygons before merging
       for lakes.

    Notes
    -------
    This function has no return values, instead will generate following
    files.
    when Is_Final_Result is true
    os.path.join(OutputFolder,'finalcat_info.shp')
    os.path.join(OutputFolder,'finalcat_info_riv.shp')
    when Is_Final_Result is False
    os.path.join(OutputFolder,'finalrivply_info.shp')
    os.path.join(OutputFolder,'finalriv_info_riv.shp')
    Returns:
    -------
    None

    Examples
    -------

    """
    maximum_obs_id = 80000
    if not os.path.exists(OutputFolder):
        os.makedirs(OutputFolder)

    tempfolder = os.path.join(
        tempfile.gettempdir(),
        "basinmaker_comsubreg" + str(np.random.randint(1, 10000 + 1)),
    )
    if not os.path.exists(tempfolder):
        os.makedirs(tempfolder)

    QgsApplication.setPrefixPath(qgis_prefix_path, True)
    Qgs = QgsApplication([], False)
    Qgs.initQgis()
    from processing.core.Processing import Processing
    from processing.tools import dataobjects
    from qgis import processing

    feedback = QgsProcessingFeedback()
    Processing.initialize()
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    context = dataobjects.createContext()
    context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)

    if not os.path.exists(OutputFolder):
        os.makedirs(OutputFolder)

    Paths_Finalcat_ply = []
    Paths_Finalcat_line = []
    Paths_Finalriv_ply = []
    Paths_Finalriv_line = []
    Paths_Con_Lake_ply = []
    Paths_None_Con_Lake_ply = []
    Paths_obs_point = []

    Path_Outlet_Down_point = subregion_inlet
    ### add new attribte
    ### find outlet subregion id
    outlet_subregion_id = Sub_Region_info.loc[
        Sub_Region_info["Dow_Sub_Reg_Id"] == maximum_obs_id - 1
    ]["Sub_Reg_ID"].values[0]
    routing_info = (
        Sub_Region_info[["Sub_Reg_ID", "Dow_Sub_Reg_Id"]].astype("float").values
    )
    Subregion_to_outlet = defcat(routing_info, outlet_subregion_id)

    ###remove subregions do not drainge to outlet subregion
    Sub_Region_info = Sub_Region_info.loc[
        Sub_Region_info["Sub_Reg_ID"].isin(Subregion_to_outlet)
    ].copy()
    routing_info = (
        Sub_Region_info[["Sub_Reg_ID", "Dow_Sub_Reg_Id"]].astype("float").values
    )

    Sub_Region_info["N_Up_SubRegion"] = np.nan
    Sub_Region_info["Outlet_SubId"] = np.nan
    subid_strat_iregion = 1
    seg_id_strat_iregion = 1
    for i in range(0, len(Sub_Region_info)):
        isubregion = Sub_Region_info["Sub_Reg_ID"].values[i]
        ProjectNM = Sub_Region_info["ProjectNM"].values[i]
        SubFolder = os.path.join(Sub_Region_OutputFolder, ProjectNM)
    
        ### define path of the output file in this sub region
        Path_Finalcat_ply = os.path.join(SubFolder, "finalcat_info.shp")
        Path_Finalcat_line = os.path.join(SubFolder, "finalcat_info_riv.shp")
        Path_Finalriv_ply = os.path.join(
            SubFolder, "catchment_without_merging_lakes.shp"
        )
        Path_Finalriv_line = os.path.join(SubFolder, "river_without_merging_lakes.shp")
        Path_Con_Lake_ply = os.path.join(SubFolder, "sl_connected_lake.shp")
        Path_None_Con_Lake_ply = os.path.join(SubFolder, "sl_non_connected_lake.shp")
        Path_obs_point = os.path.join(SubFolder, "obs_gauges.shp")
    
    
    
        ### For each subregion, add new subid to each polygon files,
        ### and append result file in the merge list
        if Is_Final_Result == True:
        ### product do not exist
            if os.path.exists(Path_Finalcat_ply) != 1:
                continue
            SubID_info = (
                Dbf_To_Dataframe(Path_Finalcat_ply)
                .drop_duplicates(subset=["SubId"], keep="first")[
                    ["SubId", "DowSubId", "Seg_ID"]
                ]
                .copy()
            )
            SubID_info = SubID_info.reset_index()
            SubID_info["nSubId"] = SubID_info.index + subid_strat_iregion + start_sub_id
            SubID_info["nSeg_ID"] = SubID_info["Seg_ID"] + seg_id_strat_iregion+ start_sub_id
    
            layer_cat = QgsVectorLayer(Path_Finalcat_ply, "")
            Add_New_SubId_To_Subregion_shpfile(
                processing,
                context,
                layer_cat,
                OutputPath=os.path.join(
                    tempfolder,
                    "finalcat_info_Region_" + str(isubregion) + "addatrri.shp",
                ),
                Region_ID=isubregion,
                SubID_info=SubID_info,
            )
            Paths_Finalcat_ply.append(
                os.path.join(
                    tempfolder,
                    "finalcat_info_Region_" + str(isubregion) + "addatrri.shp",
                )
            )
            del layer_cat
    
            layer_cat = QgsVectorLayer(Path_Finalcat_line, "")
            Add_New_SubId_To_Subregion_shpfile(
                processing,
                context,
                layer_cat,
                OutputPath=os.path.join(
                    tempfolder,
                    "finalcat_info_riv_Region_" + str(isubregion) + "addatrri.shp",
                ),
                Region_ID=isubregion,
                SubID_info=SubID_info,
            )
            Paths_Finalcat_line.append(
                os.path.join(
                    tempfolder,
                    "finalcat_info_riv_Region_" + str(isubregion) + "addatrri.shp",
                )
            )
            del layer_cat
    
        else:
            if os.path.exists(Path_Finalriv_ply) != 1:
                continue
            SubID_info = (
                Dbf_To_Dataframe(Path_Finalriv_ply)
                .drop_duplicates(subset=["SubId"], keep="first")[
                    ["SubId", "DowSubId", "Seg_ID"]
                ]
                .copy()
            )
            SubID_info = SubID_info.reset_index()
            SubID_info["nSubId"] = SubID_info.index + subid_strat_iregion + start_sub_id
            SubID_info["nSeg_ID"] = SubID_info["Seg_ID"] + seg_id_strat_iregion + start_sub_id
            layer_cat = QgsVectorLayer(Path_Finalriv_ply, "")
            Add_New_SubId_To_Subregion_shpfile(
                processing,
                context,
                layer_cat,
                OutputPath=os.path.join(
                    tempfolder,
                    "finalriv_info_ply_Region_" + str(isubregion) + "addatrri.shp",
                ),
                Region_ID=isubregion,
                SubID_info=SubID_info,
            )
            Paths_Finalriv_ply.append(
                os.path.join(
                    tempfolder,
                    "finalriv_info_ply_Region_" + str(isubregion) + "addatrri.shp",
                )
            )
            del layer_cat
    
            layer_cat = QgsVectorLayer(Path_Finalriv_line, "")
            Add_New_SubId_To_Subregion_shpfile(
                processing,
                context,
                layer_cat,
                OutputPath=os.path.join(
                    tempfolder,
                    "finalriv_info_Region_" + str(isubregion) + "addatrri.shp",
                ),
                Region_ID=isubregion,
                SubID_info=SubID_info,
            )
            Paths_Finalriv_line.append(
                os.path.join(
                    tempfolder,
                    "finalriv_info_Region_" + str(isubregion) + "addatrri.shp",
                )
            )
            del layer_cat
    
        if os.path.exists(Path_Con_Lake_ply) == 1 and os.stat(Path_Con_Lake_ply).st_size > 100:
            Paths_Con_Lake_ply.append(Path_Con_Lake_ply)
        if os.path.exists(Path_None_Con_Lake_ply) == 1 and os.stat(Path_None_Con_Lake_ply).st_size > 100:
            Paths_None_Con_Lake_ply.append(Path_None_Con_Lake_ply)
        if os.path.exists(Path_obs_point) == 1 and os.stat(Path_obs_point).st_size > 100:
            print(os.stat(Path_obs_point).st_size)
            print(isubregion)
            Paths_obs_point.append(Path_obs_point)
    
        print(
            "Subregion ID is ",
            isubregion,
            "    the start new subid is    ",
            subid_strat_iregion,
            " The end of subid is ",
            max(SubID_info["nSubId"]),
        )
    
        subid_strat_iregion = max(SubID_info["nSubId"]) + 10 - start_sub_id
        seg_id_strat_iregion = max(SubID_info["nSeg_ID"]) + 10 - start_sub_id
    

    # merge connected lake polygons
    if len(Paths_Con_Lake_ply) > 0 and not os.path.exists(os.path.join(OutputFolder, "sl_connected_lake.shp")):
        qgis_vector_merge_vector_layers(
            processing,
            context,
            INPUT_Layer_List=Paths_Con_Lake_ply,
            OUTPUT=os.path.join(OutputFolder, "sl_connected_lake.shp"),
        )
    
    # merge non connected lake polygon
    if len(Paths_None_Con_Lake_ply) > 0 and not os.path.exists(os.path.join(OutputFolder, "sl_non_connected_lake.shp")):
        qgis_vector_merge_vector_layers(
            processing,
            context,
            INPUT_Layer_List=Paths_None_Con_Lake_ply,
            OUTPUT=os.path.join(OutputFolder, "sl_non_connected_lake.shp"),
        )
    
    # merge observation points
    if len(Paths_obs_point) > 0 and not os.path.exists(os.path.join(OutputFolder, "obs_gauges.shp")):
        qgis_vector_merge_vector_layers(
            processing,
            context,
            INPUT_Layer_List=Paths_obs_point,
            OUTPUT=os.path.join(OutputFolder, "obs_gauges.shp"),
        )

    # merge catchment polygon and polyline layers, and update their attirbutes
    if Is_Final_Result == 1:
        #### Obtain downstream # id:
        qgis_vector_merge_vector_layers(
            processing,
            context,
            INPUT_Layer_List=Paths_Finalcat_ply,
            OUTPUT=os.path.join(tempfolder, "finalcat_info.shp"),
        )
        qgis_vector_merge_vector_layers(
            processing,
            context,
            INPUT_Layer_List=Paths_Finalcat_line,
            OUTPUT=os.path.join(tempfolder, "finalcat_info_riv.shp"),
        )
        processing.run(
            "qgis:joinattributesbylocation",
            {
                "INPUT": Path_Outlet_Down_point,
                "JOIN": os.path.join(tempfolder, "finalcat_info.shp"),
                "PREDICATE": [5],
                "JOIN_FIELDS": [],
                "METHOD": 1,
                "DISCARD_NONMATCHING": True,
                "PREFIX": "",
                "OUTPUT": os.path.join(tempfolder, "Down_Sub_ID.shp"),
            },
            context=context,
        )

        AllCatinfo = (
            Dbf_To_Dataframe(os.path.join(tempfolder, "finalcat_info.shp"))
            .drop_duplicates("SubId", keep="first")
            .copy()
        )

        AllCatinfo['SubId'] = AllCatinfo['SubId'].astype('int32')
        AllCatinfo['DowSubId'] = AllCatinfo['DowSubId'].astype('int32')  
        
        DownCatinfo = (
            Dbf_To_Dataframe(os.path.join(tempfolder, "Down_Sub_ID.shp"))
            .drop_duplicates("ILpt_ID", keep="first")
            .copy()
        )

        AllCatinfo, Sub_Region_info = Connect_SubRegion_Update_DownSubId(
            AllCatinfo, DownCatinfo, Sub_Region_info
        )
        AllCatinfo = Update_DA_Strahler_For_Combined_Result(AllCatinfo, Sub_Region_info,k,c)
        
        AllCatinfo['Use_region'] = 0
        Obs_names = AllCatinfo['Obs_NM'].copy(deep=True)
        Obs_names = Obs_names.fillna('nan')
        Obs_names = Obs_names.astype('string')
        mask = Obs_names == 'nan'
        AllCatinfo.loc[mask,'Has_POI'] = 0
        
        Copy_Pddataframe_to_shpfile(
            os.path.join(tempfolder, "finalcat_info.shp"),
            AllCatinfo,
            link_col_nm_shp="SubId",
            link_col_nm_df="SubId",
            UpdateColNM=["DowSubId", "DrainArea", "Strahler","DA_error","Has_POI","Use_region"],
        )
        Copy_Pddataframe_to_shpfile(
            os.path.join(tempfolder, "finalcat_info_riv.shp"),
            AllCatinfo,
            link_col_nm_shp="SubId",
            link_col_nm_df="SubId",
            UpdateColNM=["DowSubId", "DrainArea", "Strahler","DA_error","Has_POI","Use_region"],
        )
        COLUMN_NAMES_CONSTANT_Local = COLUMN_NAMES_CONSTANT_CLEAN 
        COLUMN_NAMES_CONSTANT_Local.append("Region_ID")
        COLUMN_NAMES_CONSTANT_Local.append("Use_region")
        processing.run(
            "native:dissolve",
            {
                "INPUT": os.path.join(tempfolder, "finalcat_info.shp"),
                "FIELD": ["SubId"],
                "OUTPUT": os.path.join(OutputFolder, "finalcat_info.shp"),
            },
            context=context,
        )

        Clean_Attribute_Name(
            os.path.join(OutputFolder, "finalcat_info.shp"), COLUMN_NAMES_CONSTANT_Local
        )

        processing.run(
            "native:dissolve",
            {
                "INPUT": os.path.join(tempfolder, "finalcat_info_riv.shp"),
                "FIELD": ["SubId"],
                "OUTPUT": os.path.join(OutputFolder, "finalcat_info_riv.shp"),
            },
            context=context,
        )
        Clean_Attribute_Name(
            os.path.join(OutputFolder, "finalcat_info_riv.shp"), COLUMN_NAMES_CONSTANT_Local
        )


    else:
        qgis_vector_merge_vector_layers(
            processing,
            context,
            INPUT_Layer_List=Paths_Finalriv_ply,
            OUTPUT=os.path.join(tempfolder, "finalriv_info_ply.shp"),
        )
        qgis_vector_merge_vector_layers(
            processing,
            context,
            INPUT_Layer_List=Paths_Finalriv_line,
            OUTPUT=os.path.join(tempfolder, "finalriv_info.shp"),
        )

        processing.run(
            "qgis:joinattributesbylocation",
            {
                "INPUT": Path_Outlet_Down_point,
                "JOIN": os.path.join(tempfolder, "finalriv_info_ply.shp"),
                "PREDICATE": [5],
                "JOIN_FIELDS": [],
                "METHOD": 1,
                "DISCARD_NONMATCHING": True,
                "PREFIX": "",
                "OUTPUT": os.path.join(tempfolder, "Down_Sub_ID.shp"),
            },
            context=context,
        )

        AllCatinfo = (
            Dbf_To_Dataframe(os.path.join(tempfolder, "finalriv_info_ply.shp"))
            .drop_duplicates("SubId", keep="first")
            .copy()
        )
        DownCatinfo = (
            Dbf_To_Dataframe(os.path.join(tempfolder, "Down_Sub_ID.shp"))
            .drop_duplicates("ILpt_ID", keep="first")
            .copy()
        )
        AllCatinfo['SubId'] = AllCatinfo['SubId'].astype('int32')
        AllCatinfo['DowSubId'] = AllCatinfo['DowSubId'].astype('int32')  
        
        AllCatinfo, Sub_Region_info = Connect_SubRegion_Update_DownSubId(
            AllCatinfo, DownCatinfo, Sub_Region_info
        )
        AllCatinfo['Use_region'] = 0
        AllCatinfo = Update_DA_Strahler_For_Combined_Result(AllCatinfo, Sub_Region_info,k,c)
        Obs_names = AllCatinfo['Obs_NM'].copy(deep=True)
        Obs_names = Obs_names.fillna('nan')
        Obs_names = Obs_names.astype('string')
        mask = Obs_names == 'nan'
        AllCatinfo.loc[mask,'Has_POI'] = 0
        
        DA_obs = AllCatinfo['DA_Obs'].copy(deep=True)
        DA_obs = DA_obs.fillna(0)
        DA_obs = DA_obs.astype('int32')
        mask = DA_obs <= 0
        AllCatinfo.loc[mask,'DA_error'] = 0

                
        Copy_Pddataframe_to_shpfile(
            os.path.join(tempfolder, "finalriv_info_ply.shp"),
            AllCatinfo,
            link_col_nm_shp="SubId",
            link_col_nm_df="SubId",
            UpdateColNM=["DowSubId", "DrainArea", "Strahler","DA_error","Has_POI","Use_region"],
        )
        Copy_Pddataframe_to_shpfile(
            os.path.join(tempfolder, "finalriv_info.shp"),
            AllCatinfo,
            link_col_nm_shp="SubId",
            link_col_nm_df="SubId",
            UpdateColNM=["DowSubId", "DrainArea", "Strahler","DA_error","Has_POI","Use_region"],
        )

        processing.run(
            "native:dissolve",
            {
                "INPUT": os.path.join(tempfolder, "finalriv_info_ply.shp"),
                "FIELD": ["SubId"],
                "OUTPUT": os.path.join(OutputFolder, "catchment_without_merging_lakes.shp"),
            },
            context=context,
        )
        COLUMN_NAMES_CONSTANT_Local = COLUMN_NAMES_CONSTANT 
        COLUMN_NAMES_CONSTANT_Local.append("Region_ID")
        COLUMN_NAMES_CONSTANT_Local.append("Use_region")
        Clean_Attribute_Name(
            os.path.join(OutputFolder, "catchment_without_merging_lakes.shp"), COLUMN_NAMES_CONSTANT_Local
        )

        processing.run(
            "native:dissolve",
            {
                "INPUT": os.path.join(tempfolder, "finalriv_info.shp"),
                "FIELD": ["SubId"],
                "OUTPUT": os.path.join(OutputFolder, "river_without_merging_lakes.shp"),
            },
            context=context,
        )

        Clean_Attribute_Name(
            os.path.join(OutputFolder, "river_without_merging_lakes.shp"), COLUMN_NAMES_CONSTANT_Local
        )



import numpy as np
from scipy.optimize import curve_fit
import arcpy
from arcpy import env
from arcpy.sa import *
import copy
import sys
import shutil
import os
import csv
from simpledbf import Dbf5
import pandas as pd
from shutil import copyfile
import tempfile
from basinmaker.func.pdtable import *

arcpy.env.overwriteOutput = True
arcpy.CheckOutExtension("Spatial")
##### Readed inputs

def Area_Weighted_Mapping_Between_Two_Polygons_Arcgis(Target_Ply_Path,Mapping_Ply_Path,Col_NM="HRU_ID",Output_Folder="#"):
    """Generate Grid polygon from NetCDF file
    Function that used to generate grid polygon from a NetCDF file
    Parameters
    ----------
    Target_Ply_Path                       : string
        It is the path of one inputs HRU polygon file
    Mapping_Ply_Path                      : string
        It is the path of one inputs grid polygon file
    Output_Folder                     : string
        It is the path to a folder to save output polygon shpfiles
    Notes
    -------
    Overlay_Polygons.shp                 : Polygon shpfile (output)
       It is overlay of two input polygons shpfiles
    GriddedForcings2.txt                 : Text file (output)
       It is the polygon area weighted of each polygon in Mapping_Ply_Path
       to each polygon in Target_Ply_Path
    Returns:
    -------
       None
    Examples
    -------
    """
    if not os.path.exists(Output_Folder):
        os.makedirs(Output_Folder)
        
    tempfolder = os.path.join(
        tempfile.gettempdir(), "basinmaker_" + str(np.random.randint(1, 10000 + 1))
    )
    if not os.path.exists(tempfolder):
        os.makedirs(tempfolder)

    Focspre = arcpy.Describe(Mapping_Ply_Path).spatialReference

    arcpy.Project_management(Target_Ply_Path, os.path.join(tempfolder,"temp1.shp"),Focspre)
    arcpy.RepairGeometry_management(os.path.join(tempfolder,"temp1.shp"))
    
    arcpy.Identity_analysis(os.path.join(tempfolder,"temp1.shp"), Mapping_Ply_Path,os.path.join(tempfolder,"finalcat_Frocing.shp"))
    arcpy.Dissolve_management(os.path.join(tempfolder,"finalcat_Frocing.shp"), os.path.join(Output_Folder,"Overlay_Polygons.shp"),[Col_NM, "FGID"
    ,"Row","Col"], "", "", "")
    
    if Focspre.type == "Geographic":
        arcpy.env.CoordinateSystem = arcpy.SpatialReference(3573)####wgs84 - north pore canada
    
    arcpy.AddField_management(os.path.join(Output_Folder,"Overlay_Polygons.shp"),"s_area","DOUBLE","#","#","#","#","NULLABLE","NON_REQUIRED","#")
    arcpy.CalculateField_management(os.path.join(Output_Folder,"Overlay_Polygons.shp"),"s_area","!shape.area@squaremeters!","PYTHON_9.3","#")

    dbf1 = Dbf5(Mapping_Ply_Path[:-3] + "dbf")
    Forcinfo = dbf1.to_dataframe()
    Avafgid = Forcinfo["FGID"].values

    dbf2 = Dbf5(os.path.join(Output_Folder,"Overlay_Polygons.shp")[:-3] + "dbf")
    Mapforcing = dbf2.to_dataframe()
    Mapforcing["Map_FGID"] = Mapforcing["FGID"]
    Mapforcing["Map_Row"] = Mapforcing["Row"]
    Mapforcing["Map_Col"] = Mapforcing["Col"]        
    Mapforcing = Mapforcing.loc[Mapforcing[Col_NM] > 0]  ### remove
    Mapforcing = Mapforcing.loc[Mapforcing["Map_FGID"] > 0]  ### remove
    Mapforcing = Mapforcing.loc[Mapforcing["s_area"] > 0.000001]  ### remove 
    
    grid_weight_string = create_grid_weight_main(Mapforcing,Forcinfo)
    
    ####
    grid_weight_file_path = os.path.join(Output_Folder, "GriddedForcings2.txt")
        
    WriteStringToFile(grid_weight_string, grid_weight_file_path, "w")




import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="basinmaker",
    version="3.0.1",
    author="basinmaker development team",
    author_email="m43han@uwaterloo.ca",
    description="An automated GIS toolbox for watershed delineation with lakes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dustming/basinmaker",
    packages = setuptools.find_packages(
        include=[
            "basinmaker",
            "basinmaker.*",
        ],
    ),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
)



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import sqlite3


def create_catchments_attributes_template_table(
    grassdb,
    grass_location,
    input_geo_names,
    columns,
):

    catchments = input_geo_names["catchment_without_merging_lakes"]
    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
    )
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")

    con = sqlite3.connect(
        os.path.join(grassdb, grass_location, "PERMANENT", "sqlite", "sqlite.db")
    )

    catids, temp = generate_stats_list_from_grass_raster(
        grass, mode=1, input_a=catchments
    )
    catids = np.array(catids)
    attr_template = pd.DataFrame(
        np.full((len(catids), len(columns)), -9999), columns=columns
    )

    attr_template["SubId"] = catids

    return attr_template



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import sqlite3


def modify_str_r_to_add_sub_in_head_str(
    grass,
    garray,
    con,
    str_r,
    str_v,
    connected_lake,
    cat_no_lake,
    fdr_grass,
    max_memroy,  
):

    
    
    #### find str in str_r that did not over lay with the lakes 
    
    #### overlay str and lakes
    exp = "%s = if(isnull(int(%s)),null(),%s)" % ('str_with_lake', connected_lake, str_r)
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    ### obtain connected lake ids
    str_id_with_lakes, temp = generate_stats_list_from_grass_raster(
        grass, mode=1, input_a='str_with_lake'
    )

    grass_raster_setnull_array(input = str_r,output = 'str_without_lake',values = str_id_with_lakes,grass = grass)
    
    # str_without_lakes = garray.array(mapname=str_r)
    # if (len(str_id_with_lakes) > 0):
    #     mask = np.isin(str_without_lakes, str_id_with_lakes)
    #     str_without_lakes[mask] = -9999
    # 
    # temparray = garray.array()
    # temparray[:, :] = str_without_lakes[:, :]
    # temparray.write(mapname='str_without_lake', overwrite=True)
    # grass.run_command("r.null", map='str_without_lake', setnull=[-9999, 0])
    # exp = "%s = int(%s)" % (
    #     'str_without_lake',
    #     'str_without_lake',
    # )
    # grass.run_command("r.mapcalc", expression=exp, overwrite=True)
    
    
    
    ### add new outlet only to 'str_without_lake'
    
    grass.run_command("v.extract", input=str_v, layer = 1, type = 'point',output = 'str_p',where = 'type_code < 1',overwrite=True)

    sqlstat = "SELECT cat FROM %s" % str_v
    All_ids = pd.read_sql_query(sqlstat, con)
    max_subid = np.max(All_ids['cat'].values)
    
    grass.run_command(
        "v.db.addcolumn", map='str_p', columns="newcat int"
    )
    qcol = "cat + %s" %(str(int(max_subid)))
    
    grass.run_command(
        "v.db.update", map='str_p', column="newcat", qcol=qcol
    )
    grass.run_command(
        "v.to.rast", input='str_p', output='str_p', attribute_column="newcat",use="attr", overwrite=True
    )

    # remove points that 
    exp = "%s = if(isnull(%s),null(),%s)" % (
       'str_p',
       'str_without_lake',
       'str_p',
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)
    

    grass.run_command(
        "r.grow",
        input='str_p',
        output='str_p_gr',
        radius=3,
        overwrite=True,
    )
    
    exp = "%s = if(isnull(%s),null(),%s)" % (
       'str_t',
        str_r,
       'str_p_gr',
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    exp = "%s = if(isnull(%s),%s,%s)" % (
        str_r,
       'str_t',
        str_r,
       'str_t',
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)
                    
    exp = "%s = int(%s)" % (
        str_r,
        str_r,
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    grass_raster_r_stream_basins(
        grass,
        direction=fdr_grass,
        stream=str_r,
        basins=cat_no_lake,
        memory=max_memroy,
    )
    
    return 
        
    





from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import os


def preprocessing_lake_polygon(
    mask,
    path_lakefile_in="#",
    lake_attributes=[],
    grassdb="#",
    grass_location="#",
    qgis_prefix_path="#",
    threshold_con_lake=999999999,
    threshold_non_con_lake=999999999,
    gis_platform="qgis",
    lake_name="alllake",
    lake_boundary_name="lake_boundary",
    lakes_lg_cl_thres="lakes_lg_cl_thres",
    lakes_lg_ncl_thres="lakes_lg_ncl_thres",
):

    if gis_platform == "qgis":
        assert (
            grassdb != "#"
        ), "grass database folder is needed, when gis_platform = qgis"
        assert (
            grass_location != "#"
        ), "grass location name is needed, when gis_platform = qgis"
        assert (
            qgis_prefix_path != "#"
        ), "qgis prefix path is needed, when gis_platform = qgis"
        from basinmaker.preprocessing.preinputpolygonqgis import preprocessing_input_polygon
        from basinmaker.preprocessing.reprojectandclipvectorbyplyqgis import (
            obtain_polygon_boundary,
        )
        from basinmaker.preprocessing.rasterizevectorsandloadtodbqgis import (
            rasterize_vectors_and_load_to_db,
            obtain_lake_vectors_larger_than_threstholds,
        )

    if path_lakefile_in != "#":
        preprocessing_input_polygon(
            grassdb=grassdb,
            grass_location=grass_location,
            qgis_prefix_path=qgis_prefix_path,
            mask=mask,
            path_polygon=path_lakefile_in,
            ply_name=lake_name,
            attribute_names=[lake_attributes[0]],
            raster_names=[lake_name],
        )

        # obtain lake boundary lines
        obtain_polygon_boundary(
            grassdb=grassdb,
            qgis_prefix_path=qgis_prefix_path,
            ply_path=os.path.join(grassdb, lake_name + ".shp"),
            output=os.path.join(grassdb, lake_boundary_name + ".shp"),
        )
        # obtain_lake_vectors_larger_than_threstholds(
        #     grassdb=grassdb,
        #     qgis_prefix_path=qgis_prefix_path,
        #     all_lakes=lake_name,
        #     lake_attributes=lake_attributes,
        #     threshold_con_lake=threshold_con_lake,
        #     threshold_non_con_lake=threshold_non_con_lake,
        #     lakes_lg_cl_thres=lakes_lg_cl_thres,
        #     lakes_lg_ncl_thres=lakes_lg_ncl_thres,
        # )

        rasterize_vectors_and_load_to_db(
            grassdb,
            grass_location,
            qgis_prefix_path,
            mask,
            vector_path=os.path.join(grassdb, lake_boundary_name + ".shp"),
            attribue_name=lake_attributes[0],
            raster_name=lake_boundary_name,
        )

        # rasterize_vectors_and_load_to_db(
        #     grassdb,
        #     grass_location,
        #     qgis_prefix_path,
        #     mask,
        #     vector_path=os.path.join(grassdb, lakes_lg_cl_thres + ".shp"),
        #     attribue_name=lake_attributes[0],
        #     raster_name=lakes_lg_cl_thres,
        # )
        # rasterize_vectors_and_load_to_db(
        #     grassdb,
        #     grass_location,
        #     qgis_prefix_path,
        #     mask,
        #     vector_path=os.path.join(grassdb, lakes_lg_ncl_thres + ".shp"),
        #     attribue_name=lake_attributes[0],
        #     raster_name=lakes_lg_ncl_thres,
        # )



import pandas as pd
import geopandas
from basinmaker.postprocessing.downloadpd import Download_Routing_Product_For_One_Gauge
import os
import numpy as np 

def Download_Routing_Product_From_Points_Or_LatLon(product_name,Lat = [-1],Lon = [-1]):
    
    data = pd.DataFrame(
    {
     'Latitude': Lat,
     'Longitude': Lon,
     }
     )
    
    data_gpd = geopandas.GeoDataFrame(data, geometry=geopandas.points_from_xy(data.Longitude, data.Latitude))
    data_gpd.crs = "EPSG:4326"
    
    if product_name == 'OLRP':
        version = 'v1-0'
        Drainage_region = geopandas.read_file("https://github.com/dustming/RoutingTool/wiki/Files/OLRRP_drainage_region.geojson")
    
        data_dr = geopandas.overlay(data_gpd, Drainage_region, how='identity')
        if len(data_dr) >= 1 and data_dr['Region'].values[0] in np.unique(Drainage_region['Region'].values):

            dr = data_dr['Region'].values[0]
            Subid,product_path = Download_Routing_Product_For_One_Gauge(gauge_name='#',product_name = product_name,region=dr,subreg = '#')
            product_ply_path = os.path.join(product_path,'catchment_without_merging_lakes_'+version+'.shp')
            product = geopandas.read_file(product_ply_path)
            data_gpd = data_gpd.to_crs(product.crs)
            data_pt = geopandas.overlay(data_gpd, product, how='identity')
            Subid = data_pt['SubId'].values[0]
            print("The needed product locates at:",product_path)
            print("The Subbasin Id of the lat, lon is:",Subid)
            return Subid,product_path 

        else:
            print("The point did not overlay with the routing product")
            return -1,-1

    if product_name == 'NALRP':
        version = 'v2-1'
        Drainage_region = geopandas.read_file("https://github.com/dustming/RoutingTool/wiki/Files/NA_drainage_region.geojson")
    
        data_dr = geopandas.overlay(data_gpd, Drainage_region, how='identity')
        if len(data_dr) >= 1:

            dr = data_dr['Region'].values[0]
            Subid,product_path = Download_Routing_Product_For_One_Gauge(gauge_name='#',product_name = product_name,region=dr,subreg = '#')
            product_ply_path = os.path.join(product_path,'catchment_without_merging_lakes_'+version+'.shp')
            product = geopandas.read_file(product_ply_path)
            data_gpd = data_gpd.to_crs(product.crs)
            data_pt = geopandas.overlay(data_gpd, product, how='identity')
            Subid = data_pt['SubId'].values[0]
            print("The needed product locates at:",product_path)
            print("The Subbasin Id of the lat, lon is:",Subid)
            return Subid,product_path 

        else:
            print("The point did not overlay with the routing product")
            return -1,-1
            
                        
    return -1, -1 


import copy
import numpy as np
import pandas as pd
from grass.script import array as garray

from basinmaker.func.pdtable import Check_If_Lake_Have_Multi_OutLet


####
def grass_raster_setnull(
    grass, raster_nm, null_values, create_new_raster, new_raster_nm="#"
):

    if create_new_raster:
        grass_raster_copy(grass, raster_nm, new_raster_nm)
        grass.run_command("r.null", map=new_raster_nm, setnull=null_values)
    else:
        grass.run_command("r.null", map=raster_nm, setnull=null_values)

def grass_raster_setnull_array(input,output,values,grass):
    
    raster_array = garray.array(mapname=input)
    if (len(values) > 0):
        mask = np.isin(raster_array, values)
        raster_array[mask] = -9999

    temparray = garray.array()
    temparray[:, :] = raster_array[:, :]
    temparray.write(mapname=output, overwrite=True)
    grass.run_command("r.null", map=output, setnull=[-9999, 0])
    exp = "%s = int(%s)" % (
        output,
        output,
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)
    del temparray
    del raster_array

    
#####


def grass_raster_copy(grass, raster_nm_in, raster_nm_new):
    grass.run_command("g.copy", rast=(raster_nm_in, raster_nm_new), overwrite=True)


#####


def Return_Raster_As_Array_With_Db_Path(grassdb, grass_location, raster_mn):
    """Transfer an rater in grass database into np array
    Parameters
    ----------
    grassdb         : string
    Full path to a grass database
    grass_location  : string
    location name in that grass database
    raster_mn       : string
    raster name

    Returns:
    -------
    Array            : array
    np array of the raster.

    """
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")
    Array = copy.deepcopy(garray.array(mapname=raster_mn))
    PERMANENT.close()
    Array[Array <= 0] = -9999
    return Array


###
def Return_Raster_As_Array_With_garray(garray_f, raster_mn):
    """Transfer an rater in grass database into np array
    Parameters
    ----------
    raster_mn       : string
    raster name

    Returns:
    -------
    Array            : array
    np array of the raster.

    """
    Array = copy.deepcopy(garray_f.array(mapname=raster_mn))
    Array[Array <= 0] = -9999
    return Array


def grass_raster_r_in_gdal(grass, raster_path, output_nm, location="#"):
    """import dem to target location
    Parameters
    ----------

    Returns:
    -------

    """
    if location != "#":
        grass.run_command(
            "r.in.gdal",
            input=raster_path,
            output=output_nm,
            overwrite=True,
            location=location,
        )
    else:
        grass.run_command(
            "r.in.gdal", input=raster_path, output=output_nm, overwrite=True
        )


###


def grass_raster_r_stream_extract(
    grass,
    elevation,
    accumulation,
    threshold,
    stream_raster,
    stream_vector,
    direction,
    memory,
    stream_length = -1,
):
    """using provided elevation, accumulation and threshold to derived
        stream_raster,and direction
    Parameters
    ----------

    Returns:
    -------

    """
    if stream_length < 0:
        grass.run_command(
            "r.stream.extract",
            elevation=elevation,
            accumulation=accumulation,
            threshold=threshold,
            stream_raster=stream_raster,
            direction=direction,
            stream_vector=stream_vector,
            overwrite=True,
            memory=memory,
        )
    else:
        grass.run_command(
            "r.stream.extract",
            elevation=elevation,
            accumulation=accumulation,
            threshold=threshold,
            stream_raster=stream_raster,
            direction=direction,
            stream_vector=stream_vector,
            overwrite=True,
            stream_length = stream_length,
            memory=memory,
        )
                


###


def grass_raster_r_stream_basins(grass, direction, stream, basins, memory):
    """generate catchemnt for each river segment in stream raster
    Parameters
    ----------

    Returns:
    -------

    """
    grass.run_command(
        "r.stream.basins",
        direction=direction,
        stream=stream,
        basins=basins,
        overwrite=True,
        memory=memory,
    )


###


def grass_raster_r_accumulate(grass, direction, accumulation):
    """calculate flow accumulation from flow direction dataset
    Parameters
    ----------

    Returns:
    -------

    """
    grass.run_command(
        "r.accumulate",
        direction=direction,
        accumulation=accumulation,
        overwrite=True,
    )


###


def grass_raster_r_external(grass, input, output):
    """calculate flow accumulation from flow direction dataset
    Parameters
    ----------

    Returns:
    -------

    """
    grass.run_command("r.in.gdal", input=input, output=output, overwrite=True)


###


def grass_raster_r_clip(grass, input, output):
    """clip raster with mask in grass env
    Parameters
    ----------

    Returns:
    -------

    """
    grass.run_command("r.clip", input=input, output=output, overwrite=True)


###


def grass_raster_r_to_vect(grass, input, output, type, flags="#"):
    """convert grass raster to vector
    Parameters
    ----------

    Returns:
    -------

    """
    if flags == "#":
        grass.run_command(
            "r.to.vect", input=input, output=output, type=type, overwrite=True
        )

    else:
        grass.run_command(
            "r.to.vect",
            input=input,
            output=output,
            type=type,
            flags=flags,
            overwrite=True,
        )


###


def grass_raster_r_mask(grass, raster_nm, vector_nm="#"):
    """define grass working mask for current location
    Parameters
    ----------

    Returns:
    -------

    """
    if vector_nm == "#":
        grass.run_command("r.mask", raster=raster_nm, maskcats="*", overwrite=True)
    else:
        grass.run_command("r.mask", vector=vector_nm, maskcats="*", overwrite=True)


###


def grass_raster_g_region(grass, raster_nm, zoom="#"):
    """define grass working region for current location
    Parameters
    ----------

    Returns:
    -------

    """
    if zoom == "#":
        grass.run_command("g.region", raster=raster_nm)
    else:
        grass.run_command("g.region", zoom=zoom)


###


def grass_raster_r_unpack(grass, input, output):
    """unpack pregenerated grass rasters
    Parameters
    ----------

    Returns:
    -------

    """
    grass.run_command("r.unpack", input=input, output=output, overwrite=True)


###


def grass_raster_v_unpack(grass, input, output):
    """unpack pregenerated grass vectors
    Parameters
    ----------

    Returns:
    -------

    """
    grass.run_command("v.unpack", input=input, output=output, overwrite=True)


###


def grass_raster_r_reclass(grass, input, output, rules):
    """reclassify grass raster dataset
    Parameters
    ----------

    Returns:
    -------

    """
    grass.run_command(
        "r.reclass", input=input, output=output, rules=rules, overwrite=True
    )


###


def grass_raster_r_watershed(grass, elevation, drainage, accumulation, flags):
    """generate watershed from dem,output includes flow accumulation
        and flow direction
    Parameters
    ----------

    Returns:
    -------

    """
    grass.run_command(
        "r.watershed",
        elevation=elevation,
        drainage=drainage,
        accumulation=accumulation,
        flags=flags,
        overwrite=True,
    )


###


def grass_raster_r_mapcalc(grass, expression):
    """grass map calculator
    Parameters
    ----------

    Returns:
    -------

    """
    grass.run_command("r.mapcalc", expression=expression, overwrite=True)


###


def grass_raster_create_raster_empty_raster(garray, raster_nm):
    """grass create a raster with -9999 in current location
    Parameters
    ----------

    Returns:
    -------

    """
    temparray = garray.array()
    temparray[:, :] = -9999
    temparray.write(mapname=raster_nm, overwrite=True)


###


def grass_raster_v_to_raster(grass, input, output, column, use="attr"):
    """grass create a raster with -9999 in current location
    Parameters
    ----------

    Returns:
    -------

    """
    if use == "attr":
        grass.run_command(
            "v.to.rast",
            input=input,
            output=output,
            use="attr",
            attribute_column=column,
            overwrite=True,
        )
    else:
        grass.run_command(
            "v.to.rast", input=input, output=output, use="cat", overwrite=True
        )


###


def grass_raster_r_water_outlet(
    grass, input_dir_nm, output_watshed_nm, outlet_coordinates
):
    """grass generate watershed based on outlet points
    Parameters
    ----------

    Returns:
    -------

    """
    grass.run_command(
        "r.water.outlet",
        input=input_dir_nm,
        output=output_watshed_nm,
        coordinates=outlet_coordinates,
        overwrite=True,
    )


###


def grass_raster_r_out_gdal(grass, input_nm, output, format="GTiff"):
    """grass export raster in grass working enviroment to other folder in
        different format.
    Parameters
    ----------

    Returns:
    -------

    """
    grass.run_command(
        "r.out.gdal", input=input_nm, output=output, format="GTiff", overwrite=True
    )


###


def grass_raster_r_stream_snap(
    grass, input, output, stream_rast, accumulation, radius, memory
):
    """snap observation point to closet river network
    Parameters
    ----------

    Returns:
    -------

    """
    grass.run_command(
        "r.stream.snap",
        input=input,
        output=output,
        stream_rast=stream_rast,
        accumulation=accumulation,
        radius=radius,
        overwrite=True,
        quiet="Ture",
        memory=memory,
    )


###


def DefineConnected_Non_Connected_Lakes(
    self,
    grass,
    con,
    garray,
    Routing_info,
    str_r="str_grass_r",
    Lake_r="alllake",
    Remove_Lake_Multi_OL=True,
):

    #### overlay str and lakes
    exp = "Connect_Lake_IDs = if(isnull('%s'),null(),%s)" % (str_r, Lake_r)
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    ### obtain connected lake ids
    Connect_Lake_Ids, temp = generate_stats_list_from_grass_raster(
        grass, mode=1, input_a="Connect_Lake_IDs"
    )

    #### create non connected lake raster
    grass.run_command("g.copy", rast=(Lake_r, "Nonconnect_Lake"), overwrite=True)
    grass.run_command(
        "r.null", map="Nonconnect_Lake", setnull=Connect_Lake_Ids, overwrite=True
    )
    #### create potential connected lake raster
    exp = "Connect_Lake = if(isnull('%s'),%s,null())" % ("Nonconnect_Lake", Lake_r)
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    if Remove_Lake_Multi_OL == False:
        return

    ##### obtain lake id and coresponding overlaied str id
    CL_Id, Str_Id = generate_stats_list_from_grass_raster(
        grass, mode=2, input_a="Connect_Lake", input_b=str_r
    )

    Lakes_WIth_Multi_Outlet, Remove_Str = Check_If_Lake_Have_Multi_OutLet(
        CL_Id, Str_Id, Routing_info
    )
    ### no lake has multi outlet

    if len(Lakes_WIth_Multi_Outlet) > 0:
        #        grass.run_command('r.null',map='Connect_Lake', setnull = Lakes_WIth_Multi_Outlet,overwrite = True)
        print("Following lakes have identified outlet :       ")
        print("     ", Lakes_WIth_Multi_Outlet)
    if len(Remove_Str) > 0:
        #        grass.run_command('r.null',map='str_grass_r', setnull = Remove_Str,overwrite = True)
        #        grass.run_command('r.null',map='cat1', setnull = Remove_Str,overwrite = True)
        print(
            "Following stream have been identified to make each lake has one lake outlet:       "
        )
        print("     ", Remove_Str)
    return Remove_Str


def generate_stats_list_from_grass_raster(
    grass, mode=1, input_a="P_Connect_Lake", input_b="str_grass_r", method="max"
):

    list_a = []
    list_b = []
    if mode == 1:
        p = grass.pipe_command("r.category", map=input_a, separator=" ", quiet=True)
    elif mode == 2:
        p = grass.pipe_command(
            "r.stats", input=[input_a, input_b], flags="n", quiet=True
        )
    else:
        print("error mode opton did not exist")

    for line in p.stdout:
        line_str = line.decode("utf8").rstrip("\r\n").split(" ")
        list_a.append(int(line_str[0]))
        if mode != 1:
            list_b.append(int(line_str[1]))
    p.wait()

    return list_a, list_b


def grass_raster_v_in_org(grass, input_path, output_vector_nm, location):
    """grass load vector into target grass location
    Parameters
    ----------

    Returns:
    -------

    """
    grass.run_command(
        "v.in.ogr",
        input=input_path,
        output=output_vector_nm,
        overwrite=True,
        location=location,
    )


###


def grass_raster_v_import(grass, input_path, output_vector_nm):
    """grass load vector into target grass location
    Parameters
    ----------

    Returns:
    -------

    """
    grass.run_command(
        "v.import", input=input_path, output=output_vector_nm, overwrite=True
    )


###


def grass_raster_v_db_join(grass, map, column, other_table, other_column):
    """grass join vector database
    Parameters
    ----------

    Returns:
    -------

    """
    grass.run_command(
        "v.db.join",
        map=map,
        column=column,
        other_table=other_table,
        other_column=other_column,
        overwrite=True,
    )


###
def generate_routing_info_of_catchments(
    grass, con, garray,cat="Net_cat", acc="acc_grass", Name="a1", str="#"
):

    Raster_res = grass.core.region()["nsres"]

    ###find outlet point of each cat
    grass.run_command(
        "r.stats.zonal",
        base=cat,
        cover=acc,
        method="max",
        output=Name + "_maxacc",
        overwrite=True,
    )
    ### Find the grid that equal to the max acc, thus this is the outlet grids

    exp = "%s =if(abs(%s - %s) < 1.0e-15,%s,null())" % (
        Name + "_OL1",
        acc,
        Name + "_maxacc",
        cat,
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)
    ### change outlet points to point vector
    grass.run_command(
        "r.to.vect",
        input=Name + "_OL1",
        output=Name + "_OL_v1",
        type="point",
#        flags="v",
        overwrite=True,
    )
    grass_raster_v_to_raster(
        grass,
        input=Name + "_OL_v1",
        output=Name + "_OL_2",
        column="#",
        use="cat",
    )

    grass.run_command("v.db.addcolumn", map=Name + "_OL_v1", columns="SubId int")
    grass.run_command("v.what.rast", map=Name + "_OL_v1", raster=acc, column="OL_acc")
    grass.run_command("v.what.rast", map=Name + "_OL_v1", raster=Name + "_OL1", column="SubId")

## section to remove fake outlet points, may not needed when using r.statistics in stead of
#  r.stats.zonal
###############################################################
###    # remove fake outlet,
    sqlstat = "SELECT cat, OL_acc,SubId FROM %s" % (
        Name + "_OL_v1"
    )
    outletinfo_temp = pd.read_sql_query(sqlstat, con)
    outletinfo_temp['count'] = outletinfo_temp.groupby('SubId')['SubId'].transform('count')
    outletinfo_temp['maxacc'] = outletinfo_temp.groupby('SubId')['OL_acc'].transform('max')
    outletinfo_temp = outletinfo_temp.sort_values(by='OL_acc', ascending=False)
    outletinfo_temp = outletinfo_temp.drop_duplicates(subset=['SubId'], keep='first')
    extract_cat = outletinfo_temp['cat'].values

    # for k in range(0,len(outletinfo_temp)):
    #     if outletinfo_temp['OL_acc'].values[k] == outletinfo_temp['maxacc'].values[k]:
    #          extract_cat.append(outletinfo_temp['cat'].values[k])

    # remove cat not belongs to fake outlet
    # extract_cat = np.array(extract_cat)
    array_cat_OL2 = garray.array(mapname=Name + "_OL_2")
    mask = np.logical_not(np.isin(array_cat_OL2, extract_cat))
    array_cat_OL2[mask] = -9999
    temparray = garray.array()
    temparray[:, :] = array_cat_OL2[:, :]
    temparray.write(mapname=Name + "_OL3", overwrite=True)
    grass.run_command("r.null", map=Name + "_OL3", setnull=[-9999])
    exp = "%s = int(%s)" % (
        Name + "_OL3",
        Name + "_OL3",
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    exp = "%s = if(isnull(%s),null(),%s)" % (
        Name + "_OL",
        Name + "_OL3",
        cat,
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    grass.run_command(
        "r.to.vect",
        input=Name + "_OL",
        output=Name + "_OL_v",
        type="point",
        flags="v",
        overwrite=True,
    )
    grass.run_command("v.db.addcolumn", map=Name + "_OL_v", columns="SubId int")
    grass.run_command("v.what.rast", map=Name + "_OL_v", raster=acc, column="OL_acc")
    grass.run_command("v.what.rast", map=Name + "_OL_v", raster=Name + "_OL", column="SubId")



    # grass.run_command(
    #     "v.extract",
    #     input=Name + "_OL_v1",
    #     output=Name + "_OL_v",
    #     cats=extract_cat,
    #     overwrite=True,
    # )
    # grass.run_command(
    #     "v.to.rast",
    #     input=Name + "_OL_v",
    #     output=Name + "_OL",
    #     use="attr",
    #     attribute_column='SubId',
    #     overwrite=True,
    # )
####################################################

#    grass.run_command("v.db.update", map=Name + "_OL_v", column="SubId", qcol="cat")
#    grass.run_command("v.what.rast", map=Name + "_OL_v", raster=acc, column="OL_acc")
    grass.run_command(
        "v.buffer",
        input=Name + "_OL_v",
        output=Name + "_OL_v_bf",
        distance=1.5 * Raster_res,
        flags="t",
        overwrite=True,
    )
    ###
    ###find inlet  point of each cat
    exp = "'%s' =if(isnull(%s),null(),%s)" % (Name + "_acc_riv", str, acc)
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)
    grass.run_command("r.null", map=Name + "_acc_riv", setnull=[-9999, 0])
    grass.run_command(
        "r.stats.zonal",
        base=str,
        cover=Name + "_acc_riv",
        method="min",
        output=Name + "_minacc",
        overwrite=True,
    )
    ### Find the grid that equal to the max acc, thus this is the outlet grids

    exp = "%s = int(%s)" % (
        Name + "_acc_riv",
        Name + "_acc_riv",
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    exp = "%s =if(abs(%s - %s) < 1.0e-15,%s,null())" % (
        Name + "_IL",
        Name + "_acc_riv",
        Name + "_minacc",
        str,
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)
    ### change outlet points to point vector
    grass.run_command(
        "r.to.vect",
        input=Name + "_IL",
        output=Name + "_IL_v",
        type="point",
        flags="v",
        overwrite=True,
    )
    grass.run_command("v.db.addcolumn", map=Name + "_IL_v", columns="IL_SubId int")
    #    grass.run_command('v.db.addcolumn', map=  Name+'_IL_v', columns = "IL_acc int")
    grass.run_command("v.db.update", map=Name + "_IL_v", column="IL_SubId", qcol="cat")
    grass.run_command("v.what.rast", map=Name + "_IL_v", raster=acc, column="IL_acc")
    #    grass.run_command('v.db.update', map=  Name+'_IL_v', column = "IL_acc",qcol = 'IL_acc_db - 1')

    #### add max acc inlet point acc value to Name+'_OL_v_bf'
    grass.run_command(
        "v.vect.stats",
        point=Name + "_IL_v",
        areas=Name + "_OL_v_bf",
        method="max_cat",
        points_column="IL_acc",
        count_column="IL_Pt_CT",
        stats_column="max_cat",
    )
    grass.run_command(
        "v.vect.stats",
        point=Name + "_IL_v",
        areas=Name + "_OL_v_bf",
        method="min_cat",
        points_column="IL_acc",
        count_column="IL_Pt_CT2",
        stats_column="min_cat",
    )
    grass.run_command(
        "v.vect.stats",
        point=Name + "_IL_v",
        areas=Name + "_OL_v_bf",
        method="maximum",
        points_column="IL_acc",
        count_column="IL_Pt_CT3",
        stats_column="ILmaxacc",
    )
    grass.run_command(
        "v.vect.stats",
        point=Name + "_IL_v",
        areas=Name + "_OL_v_bf",
        method="minimum",
        points_column="IL_acc",
        count_column="IL_Pt_CT4",
        stats_column="ILminacc",
    )

    #### add outlet's next inlet subid into Name+'_OL_v_bf'
    grass.run_command(
        "v.db.join",
        map=Name + "_OL_v_bf",
        column="max_cat",
        other_table=Name + "_IL_v",
        other_column="cat",
        subset_columns="IL_SubId",
        overwrite=True,
    )
    grass.run_command("v.db.addcolumn", map=Name + "_OL_v_bf", columns="ILSubIdmax int")
    grass.run_command(
        "v.db.addcolumn", map=Name + "_OL_v_bf", columns="ILAccstrmin int"
    )

    grass.run_command(
        "v.db.update", map=Name + "_OL_v_bf", column="ILSubIdmax", qcol="IL_SubId"
    )
    grass.run_command(
        "v.db.update", map=Name + "_OL_v_bf", column="ILAccstrmin", qcol="ILmaxacc"
    )
    ### add min next inlet subid to Name+'_OL_v_bf'
    grass.run_command(
        "v.db.join",
        map=Name + "_OL_v_bf",
        column="min_cat",
        other_table=Name + "_IL_v",
        other_column="cat",
        subset_columns="IL_SubId",
        overwrite=True,
    )
    grass.run_command("v.db.addcolumn", map=Name + "_OL_v_bf", columns="ILSubIdmin int")
    grass.run_command(
        "v.db.addcolumn", map=Name + "_OL_v_bf", columns="ILAccstrmax int"
    )

    grass.run_command(
        "v.db.update", map=Name + "_OL_v_bf", column="ILSubIdmin", qcol="IL_SubId"
    )
    grass.run_command(
        "v.db.update", map=Name + "_OL_v_bf", column="ILAccstrmin", qcol="ILminacc"
    )

    grass.run_command("v.db.addcolumn", map=Name + "_OL_v_bf", columns="DSubId_str int")
    grass.run_command("v.db.addcolumn", map=Name + "_OL_v_bf", columns="DSubAccstr int")
    grass.run_command(
        "v.db.update", map=Name + "_OL_v_bf", column="DSubId_str", qcol="ILSubIdmax"
    )
    grass.run_command(
        "v.db.update", map=Name + "_OL_v_bf", column="DSubAccstr", qcol="ILmaxacc"
    )

    ### in case cat drainage to itself, drainage to another close subbasin in inlet
    exp = " '%s' = '%s' " % ("SubId", "ILSubIdmax")
    grass.run_command(
        "v.db.update",
        map=Name + "_OL_v_bf",
        column="DSubId_str",
        qcol="ILSubIdmin",
        where="SubId = ILSubIdmax",
    )
    grass.run_command(
        "v.db.update",
        map=Name + "_OL_v_bf",
        column="DSubAccstr",
        qcol="ILminacc",
        where="SubId = ILSubIdmax",
    )

    ### add downsubid_Str to point vector
    grass.run_command(
        "v.db.join",
        map=Name + "_OL_v",
        column="SubId",
        other_table=Name + "_OL_v_bf",
        other_column="SubId",
        subset_columns=[
            "DSubId_str",
            "IL_Pt_CT",
            "ILSubIdmin",
            "ILSubIdmax",
            "DSubAccstr",
        ],
        overwrite=True,
    )

    ### find downn subid via cat, the reason needs to do this, is that for some cat
    ### without connected river, above approach can not find downstream cat id for
    ### these catchment

    ###add max acc in of neighbor grids as each growed grids value
    exp = "'%s'=if(isnull('%s'),null(),1)" % (Name + "_OL1", Name + "_OL")
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)
    grass.run_command(
        "r.grow",
        input=Name + "_OL1",
        output=Name + "_OL1_G",
        radius=1.5,
        overwrite=True,
    )
    # "_OL1_G_Clu" has the unique id for meeting point between catchments

    grass.run_command(
        "r.clump", input=Name + "_OL1_G", output=Name + "_OL1_G_Clu", overwrite=True
    )

    ###find inlets of each subbasin
    grass.run_command(
        "r.stats.zonal",
        base=Name + "_OL1_G_Clu",
        cover=acc,
        method="max",
        output=Name + "_OL1_G_Clu_maxacc",
        overwrite=True,
    )


    exp = "%s=if( abs(%s - %s) < 1.0e-15,%s,null())" % (
        Name + "_IL1",
        acc,
        Name + "_OL1_G_Clu_maxacc",
        cat,
    )

    grass.run_command("r.mapcalc", expression=exp, overwrite=True)
    grass.run_command(
        "r.stats.zonal",
        base=Name + "_OL1_G_Clu",
        cover=Name + "_IL1",
        method="max",
        output=Name + "_OL1_G_Clu_IL_SubId",
        overwrite=True,
    )

    grass.run_command(
        "v.what.rast",
        map=Name + "_OL_v",
        raster=Name + "_OL1_G_Clu_IL_SubId",
        column="DSubId_cat",
    )
    grass.run_command(
        "v.what.rast", map=Name + "_OL_v", raster=Name + "_maxacc", column="MaxAcc_cat"
    )
    #    grass.run_command('v.db.join', map=  Name+'_OL_v',column = 'SubId', other_table = Name+'_OL_v_bf',other_column ='SubId', subset_columns = ['DSubId_str','OL_acc','IL_Pt_CT','max_cat','IL_SubId_min'],overwrite = True)

    #### create the final downsubid use downsubid from str if it is exist
    grass.run_command("v.db.addcolumn", map=Name + "_OL_v", columns="DowSubId int")
    grass.run_command(
        "v.db.update",
        map=Name + "_OL_v",
        column="DowSubId",
        where="IL_Pt_CT > 0",
        qcol="DSubId_str",
    )
    grass.run_command(
        "v.db.update",
        map=Name + "_OL_v",
        column="DowSubId",
        where="IL_Pt_CT <= 0",
        qcol="DSubId_cat",
    )
    ### if it move to itself using dowsubid from cat
    grass.run_command(
        "v.db.update",
        map=Name + "_OL_v",
        column="DowSubId",
        where="SubId = DSubId_str",
        qcol="DSubId_cat",
    )
    ### if it move to a cat with smaller acc, use down subid from cat
    grass.run_command(
        "v.db.update",
        map=Name + "_OL_v",
        column="DowSubId",
        where="OL_acc > DSubAccstr",
        qcol="DSubId_cat",
    )
    sqlstat = "SELECT SubId, DowSubId, MaxAcc_cat FROM %s" % (Name + "_OL_v")
    Routing_info = pd.read_sql_query(sqlstat, con)

    grass.run_command(
        "r.to.vect",
        input=Name + "_IL1",
        output=Name + "_IL_v_c",
        type="point",
        overwrite=True,
    )

    grass.run_command(
        "v.what.rast",
        map=Name + "_IL_v_c",
        raster=cat,
        column="SubId_I",
    )

    grass.run_command(
        "v.what.rast",
        map=Name + "_IL_v_c",
        raster=Name + "_OL1_G_Clu",
        column="ILpt_ID",
    )

    grass.run_command(
        "v.what.rast",
        map=Name + "_OL_v",
        raster=Name + "_OL1_G_Clu",
        column="ILpt_ID",
    )
    
    un_used_map = [
                   Name + "_maxacc",
                   Name + "_OL1",
                   Name + "_OL_2",
                   Name + "_OL3",
                   Name + "_minacc",
                   Name + "_IL",
                   Name + "_OL1",
                   Name + "_OL1_G",
                   Name + "_OL1_G_Clu",
                   Name + "_OL1_G_Clu_maxacc",
                   Name + "_IL1",
                   Name + "_OL1_G_Clu_IL_SubId",
                   Name + "_IL_v_c",
    ]
    grass.run_command(
        "g.remove",
        type="raster",
        name=un_used_map,
        flags="f",
    )
    return Routing_info



import numpy as np
import sys
import os
import csv
import tempfile 
import copy 
import pandas as pd
from arcgis.features import GeoAccessor, GeoSeriesAccessor
import arcpy
from arcpy import env
from arcpy.sa import *
import shutil


import sys, os
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from basinmaker.func.arcgis import *
from basinmaker.func.pdtable import *
arcpy.env.overwriteOutput = True
arcpy.CheckOutExtension("Spatial")

def simplify_routing_structure_by_drainage_area_arcgis(
    Routing_Product_Folder= '#',
    Area_Min=-1,
    OutputFolder="#",
):
    """Simplify the routing product by drainage area

    Function that used to simplify the routing product by
    using user provided minimum subbasin drainage area.
    The input catchment polygons are routing product before
    merging for lakes. It is provided with routing product.
    The result is the simplified catchment polygons. But
    result from this fuction still not merging catchment
    covering by the same lake. Thus, The result generated
    from this tools need further processed by
    Define_Final_Catchment, or can be further processed by
    SelectLakes

    Parameters
    ----------

    Path_final_riv_ply             : string
        Path to the catchment polygon which is the routing product
        before merging lakes catchments and need to be processed before
        used. It is the input for simplify the routing product based
        on lake area or drianage area.
    Path_final_riv                 : string
        Path to the river polyline which is the routing product
        before merging lakes catchments and need to be processed before
        used. It is the input for simplify the routing product based
        on lake area or drianage area.
    Path_Con_Lake_ply              : string
        Path to a connected lake polygon. Connected lakes are lakes that
        are connected by Path_final_riv.
    Path_NonCon_Lake_ply           : string
        Path to a non connected lake polygon. Connected lakes are lakes
        that are not connected by Path_final_riv.
    Area_Min                       : float
        The minimum drainage area of each catchment in km2
    OutputFolder                   : string
        Folder name that stores generated simplified routing product

    Notes
    -------
    This function has no return values, instead will generate following
    files. The output tpye will be the same as inputs, but the routing
    network will be simplified by increase subbasin size, reduce
    number of subbasins and number of river segments.

    os.path.join(OutputFolder,os.path.basename(Path_final_riv_ply))
    os.path.join(OutputFolder,os.path.basename(Path_final_riv))
    os.path.join(OutputFolder,os.path.basename(Path_Con_Lake_ply))
    os.path.join(OutputFolder,os.path.basename(Path_NonCon_Lake_ply))

    Returns:
    -------
    None

    Examples
    -------

    """
    #### generate river catchments based on minmum area.

    sub_colnm = "SubId"
    down_colnm = "DowSubId"
    DA_colnm = "DrainArea"
    SegID_colnm = "Seg_ID"

    if not os.path.exists(OutputFolder):
        os.makedirs(OutputFolder)

    tempfolder = os.path.join(
        tempfile.gettempdir(), "basinmaker_inda_" + str(np.random.randint(1, 10000 + 1))
    )
    if not os.path.exists(tempfolder):
        os.makedirs(tempfolder)

    Path_Catchment_Polygon="#"
    Path_River_Polyline="#"
    Path_Con_Lake_ply="#"
    Path_NonCon_Lake_ply="#"
    Path_obs_gauge_point="#"
    Path_final_cat_ply="#"
    Path_final_cat_riv="#"

    ##define input files from routing prodcut 
    for file in os.listdir(Routing_Product_Folder):
        if file.endswith(".shp"):
            if 'catchment_without_merging_lakes' in file:
                Path_Catchment_Polygon = os.path.join(Routing_Product_Folder, file)
            if 'river_without_merging_lakes' in file:
                Path_River_Polyline = os.path.join(Routing_Product_Folder, file)
            if 'sl_connected_lake' in file:
                Path_Con_Lake_ply = os.path.join(Routing_Product_Folder, file)
            if 'sl_non_connected_lake' in file:
                Path_NonCon_Lake_ply = os.path.join(Routing_Product_Folder, file)
            if 'obs_gauges' in file:
                Path_obs_gauge_point = os.path.join(Routing_Product_Folder, file)
            if 'finalcat_info' in file:
                Path_final_cat_ply = os.path.join(Routing_Product_Folder, file)
            if 'finalcat_info_riv' in file:
                Path_final_cat_riv = os.path.join(Routing_Product_Folder, file)                

    if Path_Catchment_Polygon == '#' or  Path_River_Polyline =='#':
        print("Invalid routing product folder ")

    Path_final_riv_ply = Path_Catchment_Polygon
    Path_final_riv = Path_River_Polyline

    ## copy obs_gauges to output folder 
    for file in os.listdir(Routing_Product_Folder):
        if 'obs_gauges' in file:
            shutil.copy(os.path.join(Routing_Product_Folder, file), os.path.join(OutputFolder, file))


    # overall procedure,
    # 1. first get product attribute table
    # 2. determine which features needs to be merged together to increage
    #    drainage area of the sub basin, for example sub a, b c needs to be merged
    # 3. in the attribute table, change sub a b c 's content to a, assuming sub b and c drainge to a.
    # 4. copy modified attribute table to shafiles
    # 5. dissolve based on subid amd finished
    # overall procedure,

    # read attribute table, and
    Path_final_rviply = Path_final_riv_ply
    Path_final_riv = Path_final_riv
    Path_Conl_ply = Path_Con_Lake_ply
    Path_Non_ConL_ply = Path_NonCon_Lake_ply

    finalriv_infoply = pd.DataFrame.spatial.from_featureclass(Path_final_rviply)
    finalriv_inforiv = pd.DataFrame.spatial.from_featureclass(Path_final_riv)
    if Path_Conl_ply != '#':
        Conn_Lakes_ply = pd.DataFrame.spatial.from_featureclass(Path_Conl_ply)
    else:
        Conn_Lakes_ply = pd.DataFrame(np.full((10,1),-9999),columns=["Hylak_id"])

    # change attribute table
    (
        mapoldnew_info,
        Selected_riv_ids,
        Connected_Lake_Mainriv,
        Old_Non_Connect_LakeIds,
        Conn_To_NonConlakeids,
    ) = Change_Attribute_Values_For_Catchments_Need_To_Be_Merged_By_Increase_DA(
        finalriv_infoply, Conn_Lakes_ply, Area_Min
    )

    finalriv_inforiv_main = finalriv_inforiv.loc[finalriv_inforiv['SubId'].isin(Selected_riv_ids)]
    finalriv_inforiv_main.spatial.to_featureclass(location=os.path.join(tempfolder,'selected_riv.shp'),overwrite=True,sanitize_columns=False)
    
    UpdateTopology(mapoldnew_info)
    mapoldnew_info = update_non_connected_catchment_info(mapoldnew_info)

    save_modified_attributes_to_outputs(
        mapoldnew_info=mapoldnew_info,
        tempfolder=tempfolder,
        OutputFolder=OutputFolder,
        cat_name=os.path.basename(Path_final_riv_ply),
        riv_name = os.path.basename(Path_final_riv),
        Path_final_riv = os.path.join(tempfolder,'selected_riv.shp'),
    )
   
    # export lakes 
    if Path_Conl_ply == '#':
        Conn_Lakes_ply_select = []
        Conn_Lakes_ply_not_select = []
    else:     
        Conn_Lakes_ply = pd.DataFrame.spatial.from_featureclass(Path_Conl_ply)    
        lake_mask = Conn_Lakes_ply['Hylak_id'].isin(Connected_Lake_Mainriv)
        Conn_Lakes_ply_select = Conn_Lakes_ply.loc[lake_mask].copy()
        Conn_Lakes_ply_not_select = Conn_Lakes_ply.loc[np.logical_not(lake_mask)].copy()

    # export lake polygons
    # export connected lake polygon
    if len(Conn_Lakes_ply_select) > 0:
        Conn_Lakes_ply_select.spatial.to_featureclass(location=os.path.join(OutputFolder,os.path.basename(Path_Conl_ply)),overwrite=True,sanitize_columns=False)
    
    # export non connected polygon 
    if len(Conn_Lakes_ply_not_select) >0 and Path_NonCon_Lake_ply != '#':
        non_conn_Lakes_ply = pd.DataFrame.spatial.from_featureclass(Path_NonCon_Lake_ply)  
        new_non_connected_lake = pd.concat([non_conn_Lakes_ply, Conn_Lakes_ply_not_select], ignore_index=True)
        new_non_connected_lake.spatial.to_featureclass(location=os.path.join(OutputFolder,os.path.basename(Path_NonCon_Lake_ply)),overwrite=True,sanitize_columns=False)
    if len(Conn_Lakes_ply_not_select) <= 0 and Path_NonCon_Lake_ply != '#':
        non_conn_Lakes_ply = pd.DataFrame.spatial.from_featureclass(Path_NonCon_Lake_ply)
        non_conn_Lakes_ply.spatial.to_featureclass(location=os.path.join(OutputFolder,os.path.basename(Path_NonCon_Lake_ply)),overwrite=True,sanitize_columns=False)
    if len(Conn_Lakes_ply_not_select) > 0 and Path_NonCon_Lake_ply == '#':
       
        if len(os.path.basename(Path_Conl_ply).split('_')) == 4:
            outlake_name = 'sl_non_connected_lake_' + os.path.basename(Path_Conl_ply).split('_')[3]
        else:
            outlake_name = 'sl_non_connected_lake.shp'
            
        Conn_Lakes_ply_not_select.spatial.to_featureclass(location=os.path.join(OutputFolder,outlake_name),overwrite=True,sanitize_columns=False)
        
    return 



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *


def delineate_watershed_no_lake_using_fdr(
    grassdb,
    grass_location,
    qgis_prefix_path,
    input_geo_names,
    fdr_path,
    acc_thresold,
    fdr_arcgis,
    fdr_grass,
    str_r,
    str_v,
    acc,
    cat_no_lake,
    max_memroy,
):
    mask = input_geo_names["mask"]
    dem = input_geo_names["dem"]
    
    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    
    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
    )
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")

    write_grass_and_arcgis_fdr_rules(grassdb)
    print("using fdr dataset     ")
    grass_raster_r_in_gdal(grass, raster_path=fdr_path, output_nm="fdr_arcgis_temp")
    # reclassify it into grass flow direction data
    
    grass_raster_r_reclass(
        grass,
        input="fdr_arcgis_temp",
        output="fdr_grass_temp",
        rules=os.path.join(grassdb, "Arcgis2GrassDIR.txt"),
    )
    # calcuate flow accumulation from provided dir
    
    grass.run_command(
        "r.accumulate",
        direction="fdr_grass_temp",
        accumulation=acc,
        overwrite=True,
    )

    exp = "%s = %s - %s/%s" % (
        'dem_acc',
        dem,
        acc,
        str(1),
    )        
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)



    grass.run_command(
        "r.stream.extract",
        elevation='dem_acc',
        accumulation=acc,
        threshold=int(acc_thresold),
        stream_raster=str_r,
        direction=fdr_grass,
        stream_vector=str_v,
        overwrite=True,
        memory=max_memroy,
        stream_length = 10,
        d8cut = 0,
    )
                        
    # create a arcgis flow direction
    grass_raster_r_reclass(
        grass,
        input=fdr_grass,
        output=fdr_arcgis,
        rules=os.path.join(grassdb, "Grass2ArcgisDIR.txt"),
    )

    # update acc with new flow direction 
    grass_raster_r_accumulate(
        grass, direction=fdr_grass, accumulation=acc
    )
    
    
    grass_raster_r_stream_basins(
        grass,
        direction=fdr_grass,
        stream=str_r,
        basins=cat_no_lake,
        memory=max_memroy,
    )

    grass.run_command(
        "r.out.gdal",
        input=cat_no_lake,
        output=os.path.join(grassdb, "catchment_no_lake.tif"),
        format="GTiff",
        overwrite=True,
    )     
     
    PERMANENT.close()
    return



from basinmaker.utilities.utilities import Internal_Constant_Names


def add_lakes_and_obs_into_existing_watershed_delineation(
    input_geo_names,
    path_lakefile_in="#",
    lake_attributes=[],
    path_obsfile_in="#",
    obs_attributes=[],
    path_sub_reg_outlets_v="#",
    threshold_con_lake=0,
    threshold_non_con_lake=0,
    only_included_lake_at_river_interction = False,
    search_radius=100,
    sl_connected_lake="#",
    sl_non_connected_lake="#",
    sl_lakes="#",
    path_sub_reg_lake_r="#",
    path_sub_reg_lake_bd_r="#",
    catchment_without_merging_lakes="catchment_without_merging_lakes",
    river_without_merging_lakes="river_without_merging_lakes",
    snapped_obs_points="#",
    cat_use_default_acc="#",
    nfdr_arcgis="#",
    nfdr_grass="#",
    max_memroy=1024 * 4,
    grassdb="#",
    grass_location="#",
    qgis_prefix_path="#",
    gis_platform="qgis",
):
    """Add lake inflow and outflow points as new subabsin outlet

    Update the subbasin delineation result generated by
    "watershed_delineation_without_lake_method". Lake's inflow and outflow
    points will be added into subbasin delineation result as a new subbasin
    outlet. Result from this tool is not the final delineation result.
    Because some lake may cover several subbasins. these subbasin
    are not megered into one subbasin yet.

    Parameters
    ----------
    input_geo_names             : dict
        it is a dictionary that list the required input file names,should at
        least indicate the name of following items:
        fdr_arcgis                   : raster
            it is a raster represent flow direction dataset, which is
            using 1,2,4,...64,128 to represent different directions
        fdr_grass                    : raster
            it is a raster represent flow direction dataset, which is
            using 1 - 8 to represent different directions
        str_r                        : raster
            it is a river network in vector format

        str_v                        : vector
            it is a river network in raster format

        cat_no_lake                  : raster
             it is the raster represent the delineated subbasins without
             considering lakes
        mask                         : raster
            it is a mask raster stored in grass database, which indicate
            the PSE
        dem                          : raster
            it is a dem raster stored in grass database, which is
            has the same extent with MASK.
        acc                          : raster
            it is the raster represent the flow accumulation
    path_lakefile_in            : string (optional)
        It is a string to indicate the full path of the polygon
        shapefile that include lake data.
    lake_attributes             : list
        the columns names that indicate following items has to be included
        1) column name for the unique Id of each lake within the lake polygon shpfile;
        2) column name for type of the lake should be integer;
        3) column name for the volume of the lake in km3;
        4) column name for the average depth of the lake in m;
        5) column name for the area of the lake in km2.
    threshold_con_lake          : float
        It is a lake area thresthold for connected lakes in km2
        Connected Lake with lake area below this value will be
        not added into the subbasin delineation
    threshold_non_con_lake      : float
        It is a lake area thresthold for non-connected lakes in km2
        Non connected Lake with lake area below this value will be
        not added into the subbasin delineation
    path_obsfile_in             : string
        It is a string to indicate the full path of the point
        shapefile that indicate observation gauges
    obs_attributes              : list
        the columns names that indicate following items has to be included
        1) column name for the unique Id of each observation point;
        2) column name for the unique name of each observation point;
        3) column name for the drainage area of each observation point in km3;
        4) column name for the source of the observation point:
            'CA' for observation in canada;
            'US' for observation in US;
    search_radius               : integer
        It is the search ratio in number of grids to snap observation
        point into the river network.
    path_sub_reg_outlets_v      : path
        It is the full path of the subregion outlet vector file
    max_memroy                  : integer
        It is the maximum memeory that allow to be used.
    grassdb                     : path (required)
        It is a path to project grass database folder
    grass_location              : string (required)
        It is a string of grass location name
    qgis_prefix_path            : string (required)
        It is a string of qgis prefix path
    gis_platform                : string
        It is a string indicate with gis platform is used:
        'qgis'                       : the basinmaker is running within QGIS
        'arcgis'                     : the basinmaker is running within ArcGIS
    Notes
    -------
    Raster and vector files that are generated by this function and
    will be used by next step are list as following. All files are
    stored at a grass database

    selected_lakes              : raster
        it is a raster represent all lakes that are selected by two lake
        area threstholds
    sl_nonconnect_lake          : raster
        it is a raster represent all non connected lakes that are selected
        by lake area threstholds
    sl_connected_lake           : raster
        it is a raster represent allconnected lakes that are selected
        by lake area threstholds
    river_without_merging_lakes : raster/vector
        it is the updated river segment for each subbasin
    catchment_without_merging_lakes   : raster/vector
        it is a raster represent updated subbasins after adding lake inflow
        and outflow points as new subbasin outlet.
    snapped_obs_points          : raster/vector
        it is a name of the point gis file represent successfully sanpped
        observation points
    Returns:
    -------
       None

    Examples
    -------

    """

    # define internal file names
    cat_add_lake = Internal_Constant_Names["cat_add_lake"]
    pourpoints_with_lakes = Internal_Constant_Names["pourpoints_with_lakes"]
    pourpoints_add_obs = Internal_Constant_Names["pourpoints_add_obs"]
    lake_outflow_pourpoints = Internal_Constant_Names["lake_outflow_pourpoints"]

    if gis_platform == "qgis":
        assert (
            grassdb != "#"
        ), "grass database folder is needed, when gis_platform = qgis"
        assert (
            grass_location != "#"
        ), "grass location name is needed, when gis_platform = qgis"
        assert (
            qgis_prefix_path != "#"
        ), "qgis prefix path is needed, when gis_platform = qgis"
        from basinmaker.addlakeandobs.addlakesqgis import (
            add_lakes_into_existing_watershed_delineation,
        )
        from basinmaker.addlakeandobs.addobsqgis import add_obs_into_existing_watershed_delineation
        from basinmaker.addlakeandobs.definecatrivqgis import (
            define_cat_and_riv_without_merge_lake_cats,
        )
    elif gis_platform == "arcgis":
        from basinmaker.addlakeandobs.addlakesarcgis import (
            add_lakes_into_existing_watershed_delineation,
        )
    else:
        print('gis_platform    ',gis_platform,'   do not support') 
            
    lake_outflow_pourpoints = add_lakes_into_existing_watershed_delineation(
        grassdb=grassdb,
        grass_location=grass_location,
        qgis_prefix_path=qgis_prefix_path,
        input_geo_names=input_geo_names,
        path_lakefile_in=path_lakefile_in,
        lake_attributes=lake_attributes,
        threshold_con_lake=threshold_con_lake,
        path_sub_reg_lake_r=path_sub_reg_lake_r,
        path_sub_reg_lake_bd_r=path_sub_reg_lake_bd_r,
        threshold_non_con_lake=threshold_non_con_lake,
        only_included_lake_at_river_interction = only_included_lake_at_river_interction,
        sl_connected_lake=sl_connected_lake,
        sl_non_connected_lake=sl_non_connected_lake,
        sl_lakes=sl_lakes,
        nfdr_arcgis=nfdr_arcgis,
        nfdr_grass=nfdr_grass,
        cat_add_lake=cat_add_lake,
        pourpoints_with_lakes=pourpoints_with_lakes,
        cat_use_default_acc=cat_use_default_acc,
        lake_outflow_pourpoints=lake_outflow_pourpoints,
        max_memroy=max_memroy,
    )
    ## add variable into input geo names
    input_geo_names["cat_add_lake"] = cat_add_lake
    input_geo_names["pourpoints_with_lakes"] = pourpoints_with_lakes
    input_geo_names["lake_outflow_pourpoints"] = lake_outflow_pourpoints
    input_geo_names["sl_non_connected_lake"] = sl_non_connected_lake
    input_geo_names["nfdr_grass"] = nfdr_grass

    if path_obsfile_in != "#":
        add_obs_into_existing_watershed_delineation(
            grassdb=grassdb,
            grass_location=grass_location,
            qgis_prefix_path=qgis_prefix_path,
            input_geo_names=input_geo_names,
            path_obsfile_in=path_obsfile_in,
            obs_attributes=obs_attributes,
            search_radius=search_radius,
            path_sub_reg_outlets_v=path_sub_reg_outlets_v,
            max_memroy=max_memroy,
            pourpoints_add_obs=pourpoints_add_obs,
            snapped_obs_points=snapped_obs_points,
        )
    else:
        pourpoints_add_obs = pourpoints_with_lakes

    input_geo_names["pourpoints_add_obs"] = pourpoints_add_obs

    define_cat_and_riv_without_merge_lake_cats(
        grassdb=grassdb,
        grass_location=grass_location,
        qgis_prefix_path=qgis_prefix_path,
        input_geo_names=input_geo_names,
        path_lakefile_in=path_lakefile_in,
        catchment_without_merging_lakes=catchment_without_merging_lakes,
        river_without_merging_lakes=river_without_merging_lakes,
        max_memroy=max_memroy,
    )
